"""Form utility for editing field options.

The :class:`OptionsForm` class introspects the ``option_model`` dataclass
associated with a :class:`~pysigil.settings_metadata.FieldType` and renders
appropriate editor widgets for each field.  The widget layer itself is kept
very small and reuses the generic editor widgets exposed by
:mod:`pysigil.ui.widgets`.

Only a subset of basic field types is currently supported: ``str``, ``int``,
``float``, ``bool`` and ``list[str]``.  For more elaborate editors a
:class:`~pysigil.settings_metadata.FieldType` can supply its own ``option_widget``
which is used instead of the automatic dataclass based rendering.
"""

from __future__ import annotations

from dataclasses import asdict, fields, is_dataclass
from typing import Any, Mapping, get_args, get_origin, get_type_hints, Union
from types import UnionType

from ..settings_metadata import TYPE_REGISTRY, FieldType
from .widgets import EditorWidget

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk  # type: ignore
    from tkinter import ttk  # type: ignore
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    ttk = None  # type: ignore


class OptionsForm(ttk.Frame):
    """Render editors for a field type's option model."""

    def __init__(self, master, field_type: str | FieldType) -> None:
        if tk is None or ttk is None:  # pragma: no cover - tkinter missing
            raise RuntimeError("tkinter is required for OptionsForm")
        super().__init__(master, style="CardBody.TFrame")
        self._ft = TYPE_REGISTRY[field_type] if isinstance(field_type, str) else field_type
        self._option_model = self._ft.option_model
        self._widgets: dict[str, EditorWidget] = {}
        self._fields: dict[str, tuple[str, bool]] = {}
        self._custom: EditorWidget | None = None
        if self._ft.option_widget is not None:
            self._custom = self._ft.option_widget(self)
        elif self._option_model is not None and is_dataclass(self._option_model):
            self._build_from_model()
        # when there is neither a custom widget nor option model the form stays empty

    # ------------------------------------------------------------------
    def _build_from_model(self) -> None:
        assert self._option_model is not None
        hints = get_type_hints(self._option_model, include_extras=True)
        for row, f in enumerate(fields(self._option_model)):
            tp = hints.get(f.name, f.type)
            key, optional = self._field_key(tp)
            ft = TYPE_REGISTRY[key]
            if ft.value_widget is None:
                raise TypeError(f"no widget for option field type {key}")
            widget = ft.value_widget(self)  # type: ignore[assignment]
            label = ttk.Label(self, text=f.name.replace("_", " "), style="Card.TLabel")
            label.grid(row=row, column=0, sticky="w", padx=4, pady=2)
            widget.grid(row=row, column=1, sticky="ew", padx=4, pady=2)
            self._widgets[f.name] = widget
            self._fields[f.name] = (key, optional)
        self.columnconfigure(1, weight=1)

    # ------------------------------------------------------------------
    @staticmethod
    def _unwrap_optional(tp: Any) -> tuple[bool, Any]:
        origin = get_origin(tp)
        if origin in (Union, UnionType):
            args = [a for a in get_args(tp) if a is not type(None)]
            if len(args) == 1:
                return True, args[0]
        return False, tp

    def _field_key(self, tp: Any) -> tuple[str, bool]:
        optional, base = self._unwrap_optional(tp)
        origin = get_origin(base)
        if origin is list and get_args(base) == (str,):
            return "string_list", optional
        mapping = {str: "string", int: "integer", float: "number", bool: "boolean"}
        key = mapping.get(base)
        if key is None:
            raise TypeError(f"unsupported option field type: {tp!r}")
        return key, optional

    def _parse_value(self, key: str, optional: bool, raw: Any) -> Any:
        if raw in (None, "") and optional:
            return None
        ft = TYPE_REGISTRY[key]
        to_parse = None if raw in (None, "") else raw
        if to_parse is not None and not isinstance(to_parse, str):
            to_parse = ft.adapter.serialize(to_parse)
        return ft.adapter.parse(to_parse)

    def _format_value(self, key: str, value: Any) -> Any:
        if value is None:
            return None
        ft = TYPE_REGISTRY[key]
        serialized = ft.adapter.serialize(value)
        return ft.adapter.parse(serialized) if key == "boolean" else serialized

    # ------------------------------------------------------------------
    def get_values(self) -> dict[str, Any]:
        if self._custom is not None:
            val = self._custom.get_value()
            return val if isinstance(val, dict) else {}
        values: dict[str, Any] = {}
        for name, widget in self._widgets.items():
            key, optional = self._fields[name]
            raw = widget.get_value()
            values[name] = self._parse_value(key, optional, raw)
        return values

    def set_values(self, values: Mapping[str, Any] | Any | None) -> None:
        if self._custom is not None:
            self._custom.set_value({} if values is None else (asdict(values) if is_dataclass(values) else values))
            return
        if values is None:
            values = {}
        if is_dataclass(values):
            values = asdict(values)
        for name, widget in self._widgets.items():
            key, _ = self._fields[name]
            widget.set_value(self._format_value(key, values.get(name)))

    def validate(self) -> dict[str, str | None]:
        if self._custom is not None:
            # delegate validation to dataclass if available
            data = self._custom.get_value() or {}
            errors: dict[str, str | None] = {}
            if self._option_model is not None:
                try:
                    self._option_model(**data)
                except Exception as exc:  # pragma: no cover - defensive
                    errors["__all__"] = str(exc)
                else:
                    errors["__all__"] = None
            return errors
        errors: dict[str, str | None] = {}
        data: dict[str, Any] = {}
        for name, widget in self._widgets.items():
            key, optional = self._fields[name]
            raw = widget.get_value()
            try:
                value = self._parse_value(key, optional, raw)
            except Exception as exc:
                widget.set_error(str(exc))
                errors[name] = str(exc)
            else:
                widget.set_error(None)
                data[name] = value
                errors[name] = None
        if errors and any(v is not None for v in errors.values()):
            return errors
        if self._option_model is not None:
            try:
                self._option_model(**data)
            except Exception as exc:  # pragma: no cover - defensive
                errors["__all__"] = str(exc)
            else:
                errors["__all__"] = None
        return errors


__all__ = ["OptionsForm"]



"""Authoring tools UI."""

from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any, Callable

try:  # pragma: no cover - tkinter availability depends on environment
    import tkinter as tk
    from tkinter import messagebox, ttk
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    ttk = None  # type: ignore

from ...settings_metadata import TYPE_REGISTRY, FieldType, SHORT_DESC_MAX
from ..author_adapter import AuthorAdapter, FieldInfo
from ..options_form import OptionsForm
from ..core import AppCore
from ..value_parser import parse_field_value
from ..aurelia_theme import get_palette, use


class UnsavedChangesDialog(tk.Toplevel):
    """Modal dialog prompting the user about unsaved changes."""

    def __init__(self, master: tk.Widget) -> None:  # pragma: no cover - simple modal
        super().__init__(master)
        self.title("Unsaved changes")
        self.transient(master)
        self.grab_set()
        self.resizable(False, False)
        palette = get_palette()
        self.configure(bg=palette["bg"])  # type: ignore[call-arg]
        self._result: str | None = None

        body = ttk.Frame(self, padding=18)
        body.pack(fill="both", expand=True)
        ttk.Label(
            body,
            text="You have unsaved changes.",
            style="Title.TLabel",
            anchor="w",
        ).pack(fill="x")
        ttk.Label(
            body,
            text="Save before leaving?",
            anchor="w",
            padding=(0, 6, 0, 12),
        ).pack(fill="x")

        buttons = ttk.Frame(body)
        buttons.pack(fill="x", pady=(6, 0))
        ttk.Button(buttons, text="Save", command=self._on_save).pack(side="left")
        ttk.Button(buttons, text="Discard", command=self._on_discard).pack(side="left", padx=6)
        ttk.Button(buttons, text="Cancel", command=self._on_cancel).pack(side="right")

        self.protocol("WM_DELETE_WINDOW", self._on_cancel)
        self.bind("<Escape>", lambda _e: self._on_cancel())
        self.after(0, self._focus_default)

    def _focus_default(self) -> None:
        try:
            self.focus_force()
        except Exception:
            pass

    def _finish(self, result: str) -> None:
        self._result = result
        self.destroy()

    def _on_save(self) -> None:
        self._finish("save")

    def _on_discard(self) -> None:
        self._finish("discard")

    def _on_cancel(self) -> None:
        self._finish("cancel")

    def show(self) -> str | None:
        """Block until the dialog is dismissed and return the chosen action."""

        self.wait_visibility()
        self.wait_window()
        return self._result


class AuthorTools(tk.Toplevel):  # pragma: no cover - simple UI wrapper
    """Toplevel window exposing authoring helpers for provider authors."""

    def __init__(self, master: tk.Misc, core: AppCore) -> None:
        super().__init__(master)
        use(self)
        palette = get_palette()
        self.configure(bg=palette["bg"])  # type: ignore[call-arg]
        style = ttk.Style(self)
        style.configure(
            "Treeview",
            background=palette["card"],
            fieldbackground=palette["card"],
            foreground=palette["ink"],
            bordercolor=palette["card_edge"],
        )
        style.map(
            "Treeview",
            background=[("selected", palette["primary_hover"])],
            foreground=[("selected", palette["on_primary"])],
        )
        style.configure(
            "TLabel",
            background=palette["bg"],
            foreground=palette["hdr_fg"],
            font=(None, 10, "bold"),
        )

        pid = core.state.provider_id or ""
        project_root: Path | None = core.state.project_root
        project = project_root.as_posix() if project_root else ""
        title = "Sigil – Author Tools"
        if pid:
            title += f" – {pid}"
        if project:
            title += f" – {project}"
        self._base_title = title
        self.core = core
        self._info_var = tk.StringVar()
        info = f"Provider: {pid}"
        if project:
            info += f" – {project}"
        self._info_var.set(info)
        self.adapter = AuthorAdapter(pid or None)
        self._current_key: str | None = None
        self._value_widget: object | None = None
        self._options_widget: object | None = None
        self._opts_frame = None
        self._opts_container = None
        self._opts_pack: dict[str, Any] = {"fill": "x", "pady": (0, 6)}
        self._default_frame = None
        self._default_container = None
        self._default_pack: dict[str, Any] = {"fill": "x", "pady": (0, 6)}
        # Undiscovered fields are loaded on demand
        self._undiscovered_loaded = False
        self._dirty_tabs: dict[str, bool] = {"fields": False, "defaults": False, "untracked": False}
        self._window_dirty = False
        self._field_snapshot: dict[str, Any] | None = None
        self._current_tab = "fields"
        self._suspend_dirty = 0
        self._var_traces: list[tuple[tk.Variable, str]] = []
        self._widget_binds: list[tuple[tk.Widget, str, str]] = []
        self._suspend_tree_select = False
        self._current_tree_selection: str | None = None
        self.protocol("WM_DELETE_WINDOW", self._on_close_request)
        self._update_window_title()
        self._build()
        if pid:
            self._reload_tree()

    # ------------------------------------------------------------------
    # UI construction
    # ------------------------------------------------------------------
    def _build(self) -> None:
        self.geometry("800x600")
        ttk.Label(self, textvariable=self._info_var).pack(anchor="w", padx=6, pady=6)
        pw = ttk.PanedWindow(self, orient="horizontal")
        self._left = ttk.Frame(pw, style="TFrame")
        self._right = ttk.Frame(pw, style="TFrame")
        pw.add(self._left, weight=1)
        pw.add(self._right, weight=3)
        pw.pack(fill="both", expand=True)

        # -- left: search + tree -------------------------------------------------
        search = ttk.Frame(self._left, style="CardBody.TFrame")
        search.pack(fill="x", padx=6, pady=(6, 0))
        self._search_var = tk.StringVar()
        entry = ttk.Entry(search, textvariable=self._search_var)
        entry.pack(fill="x", side="left", expand=True)
        ttk.Button(search, text="Add", command=self._on_add).pack(side="right", padx=(4, 0))
        self._search_var.trace_add("write", lambda *_: self._reload_tree())

        self._tree = ttk.Treeview(self._left, show="tree")
        self._tree.pack(fill="both", expand=True, padx=6, pady=6)
        self._tree.bind("<<TreeviewSelect>>", self._on_select)
        self._tree.bind("<<TreeviewOpen>>", self._on_tree_open)

        # -- right: placeholder frame for form -----------------------------------
        self._form = ttk.Frame(self._right, style="TFrame")
        self._form.pack(fill="both", expand=True, padx=6, pady=6)

    # ------------------------------------------------------------------
    def _update_window_title(self) -> None:
        title = self._base_title
        dirty = any(self._dirty_tabs.values())
        if dirty:
            title = f"* {title}"
        try:
            self.title(title)
        except Exception:
            pass
        self._window_dirty = dirty

    def _set_dirty(self, tab: str, flag: bool) -> None:
        if self._dirty_tabs.get(tab) == flag:
            return
        self._dirty_tabs[tab] = flag
        self._update_window_title()

    def _dirty_any(self) -> bool:
        return any(self._dirty_tabs.values())

    def _clear_dirty_watchers(self) -> None:
        for var, trace_id in self._var_traces:
            try:
                var.trace_remove("write", trace_id)
            except Exception:
                pass
        self._var_traces.clear()
        for widget, sequence, bind_id in self._widget_binds:
            try:
                widget.unbind(sequence, bind_id)
            except Exception:
                pass
        self._widget_binds.clear()

    def _register_var_trace(self, var: tk.Variable) -> None:
        trace_id = var.trace_add("write", self._on_form_var_change)
        self._var_traces.append((var, trace_id))

    def _bind_widget_event(self, widget: tk.Widget, sequence: str) -> None:
        bind_id = widget.bind(sequence, self._on_widget_event, add=True)
        if bind_id:
            self._widget_binds.append((widget, sequence, bind_id))

    def _watch_widget(self, widget: tk.Widget) -> None:
        for seq in ("<KeyRelease>", "<ButtonRelease>", "<<ComboboxSelected>>", "<<ListboxSelect>>", "<<ListChanged>>"):
            self._bind_widget_event(widget, seq)
        for child in widget.winfo_children():
            if isinstance(child, tk.Widget):
                self._watch_widget(child)

    def _register_text_widget(self, widget: tk.Text) -> None:
        widget.edit_modified(False)
        bind_id = widget.bind("<<Modified>>", self._on_text_modified, add=True)
        if bind_id:
            self._widget_binds.append((widget, "<<Modified>>", bind_id))

    def _register_form_watchers(self) -> None:
        if tk is None:
            return
        self._clear_dirty_watchers()
        vars_to_watch = [
            getattr(self, "_key_var", None),
            getattr(self, "_label_var", None),
            getattr(self, "_desc_short_var", None),
            getattr(self, "_type_var", None),
            getattr(self, "_section_var", None),
            getattr(self, "_order_var", None),
        ]
        for var in vars_to_watch:
            if isinstance(var, tk.Variable):
                self._register_var_trace(var)
        desc = getattr(self, "_desc_text", None)
        if isinstance(desc, tk.Text):
            self._register_text_widget(desc)
        for widget in (self._value_widget, self._options_widget, self._form):
            if isinstance(widget, tk.Widget):
                self._watch_widget(widget)

    def _capture_form_state(self) -> dict[str, Any]:
        state: dict[str, Any] = {}
        if not hasattr(self, "_key_var"):
            return state
        state["current_key"] = self._current_key
        state["key"] = self._key_var.get() if hasattr(self, "_key_var") else ""
        state["label"] = self._label_var.get() if hasattr(self, "_label_var") else ""
        state["description_short"] = (
            self._desc_short_var.get() if hasattr(self, "_desc_short_var") else ""
        )
        if isinstance(getattr(self, "_desc_text", None), tk.Text):
            desc_widget: tk.Text = self._desc_text  # type: ignore[assignment]
            state["description"] = desc_widget.get("1.0", "end-1c")
        else:
            state["description"] = ""
        state["type"] = self._type_var.get() if hasattr(self, "_type_var") else ""
        state["section"] = self._section_var.get() if hasattr(self, "_section_var") else ""
        state["order"] = self._order_var.get() if hasattr(self, "_order_var") else ""
        default: Any | None = None
        if self._value_widget is not None and hasattr(self._value_widget, "get_value"):
            try:
                default = self._value_widget.get_value()  # type: ignore[attr-defined]
            except Exception:
                default = None
        state["default"] = deepcopy(default)
        options = self._collect_options()
        state["options"] = deepcopy(options) if options is not None else None
        return state

    def _store_field_snapshot(self) -> None:
        if not self._form.winfo_children():
            self._field_snapshot = None
            self._set_dirty("fields", False)
            return
        self._field_snapshot = deepcopy(self._capture_form_state())
        self._set_dirty("fields", False)

    def _apply_form_state(self, state: dict[str, Any] | None) -> None:
        if state is None:
            self._clear_form()
            return
        self._suspend_dirty += 1
        try:
            if hasattr(self, "_key_var"):
                self._key_var.set(state.get("key", ""))
            if hasattr(self, "_label_var"):
                self._label_var.set(state.get("label", ""))
            if hasattr(self, "_desc_short_var"):
                self._desc_short_var.set(state.get("description_short", ""))
            if isinstance(getattr(self, "_desc_text", None), tk.Text):
                desc_widget = self._desc_text  # type: ignore[assignment]
                desc_widget.delete("1.0", "end")
                desc_widget.insert("1.0", state.get("description", ""))
            if hasattr(self, "_section_var"):
                self._section_var.set(state.get("section", ""))
            if hasattr(self, "_order_var"):
                self._order_var.set(state.get("order", ""))
            type_name = state.get("type", "")
            if hasattr(self, "_type_var"):
                self._type_var.set(type_name)
            ft = TYPE_REGISTRY.get(type_name)
            self._build_type_section(ft, state.get("default"), state.get("options"))
        finally:
            self._suspend_dirty = max(0, self._suspend_dirty - 1)
        self._register_form_watchers()
        self._store_field_snapshot()

    def _update_dirty_state(self) -> None:
        if self._suspend_dirty or self._current_tab != "fields":
            return
        if self._field_snapshot is None:
            self._set_dirty("fields", bool(self._form.winfo_children()))
            return
        current = self._capture_form_state()
        self._set_dirty("fields", current != self._field_snapshot)

    def _on_form_var_change(self, *_args: object) -> None:
        if self._suspend_dirty:
            return
        self._update_dirty_state()

    def _on_widget_event(self, _event: object | None = None) -> None:
        if self._suspend_dirty:
            return
        self._update_dirty_state()

    def _on_text_modified(self, event: tk.Event) -> None:  # type: ignore[override]
        widget = event.widget
        try:
            widget.edit_modified(False)  # type: ignore[call-arg]
        except Exception:
            pass
        self._on_widget_event()

    def _on_close_request(self) -> None:
        self._confirm_unsaved(self.destroy)

    def _confirm_unsaved(
        self,
        on_continue: Callable[[], None],
        *,
        on_cancel: Callable[[], None] | None = None,
    ) -> None:
        if not self._dirty_any():
            on_continue()
            return
        dialog = UnsavedChangesDialog(self)
        choice = dialog.show() or "cancel"
        if choice == "save":
            if self._commit_current_tab():
                on_continue()
            else:
                if on_cancel is not None:
                    on_cancel()
        elif choice == "discard":
            self._discard_current_tab()
            on_continue()
        else:
            if on_cancel is not None:
                on_cancel()

    def _commit_current_tab(self) -> bool:
        if self._current_tab == "fields":
            return self._try_save_field()
        return True

    def _discard_current_tab(self) -> None:
        if self._current_tab == "fields":
            if self._field_snapshot is not None:
                self._apply_form_state(self._field_snapshot)
            else:
                self._clear_form()

    # ------------------------------------------------------------------
    def _clear_form(self) -> None:
        self._clear_dirty_watchers()
        for child in self._form.winfo_children():
            child.destroy()
        self._value_widget = None
        self._options_widget = None
        self._opts_frame = None
        self._opts_container = None
        self._opts_pack = {"fill": "x", "pady": (0, 6)}
        self._default_frame = None
        self._default_container = None
        self._default_pack = {"fill": "x", "pady": (0, 6)}
        self._field_snapshot = None
        self._set_dirty("fields", False)

    # ------------------------------------------------------------------
    def _reload_tree(self) -> None:
        """Populate tree with defined fields and a collapsible undiscovered section."""

        pattern = self._search_var.get().strip().lower()
        current = self._current_tree_selection
        self._tree.delete(*self._tree.get_children(""))

        try:
            defined = list(self.adapter.list_defined())
            undiscovered = list(self.adapter.list_undiscovered())
        except RuntimeError:
            return

        # Defined fields appear directly at the root
        for info in defined:
            if pattern and pattern not in info.key.lower():
                continue
            self._tree.insert("", "end", text=info.key, iid=f"defined:{info.key}")

        # Undiscovered fields live under a lazily populated node at the bottom
        if undiscovered:
            undis_id = self._tree.insert(
                "", "end", text="Undiscovered", iid="undiscovered", open=self._undiscovered_loaded
            )
            if self._undiscovered_loaded:
                for info in undiscovered:
                    if pattern and pattern not in info.key.lower():
                        continue
                    self._tree.insert(
                        undis_id, "end", text=info.key, iid=f"undiscovered:{info.key}"
                    )
            else:
                # insert a placeholder so the node is expandable before loading
                self._tree.insert(undis_id, "end")

        else:
            # Reset flag so the node stays collapsed if undiscovered fields later appear
            self._undiscovered_loaded = False

        if current and self._tree.exists(current):
            self._select_tree_iid(current)
        else:
            if not self._tree.selection():
                self._current_tree_selection = None

    # ------------------------------------------------------------------
    def _select_tree_iid(self, iid: str) -> None:
        if not self._tree.exists(iid):
            return
        self._suspend_tree_select = True
        try:
            self._tree.selection_set(iid)
            self._tree.focus(iid)
        finally:
            self._suspend_tree_select = False
        self._current_tree_selection = iid

    def _restore_tree_selection(self) -> None:
        target = self._current_tree_selection
        self._suspend_tree_select = True
        try:
            if target and self._tree.exists(target):
                self._tree.selection_set(target)
                self._tree.focus(target)
            else:
                self._tree.selection_remove(self._tree.selection())
        finally:
            self._suspend_tree_select = False

    # ------------------------------------------------------------------
    def _on_tree_open(self, _event: object | None = None) -> None:
        node = self._tree.focus()
        if node == "undiscovered" and not self._undiscovered_loaded:
            self._undiscovered_loaded = True
            self._reload_tree()

    # ------------------------------------------------------------------
    def _build_type_section(
        self,
        field_type: FieldType | None,
        default: object | None = None,
        options: object | None = None,
    ) -> None:
        if self._opts_frame is None or self._default_frame is None:
            return
        for child in self._opts_frame.winfo_children():
            child.destroy()
        for child in self._default_frame.winfo_children():
            child.destroy()
        self._value_widget = None
        self._options_widget = None
        if field_type is None:
            return
        # Options
        if field_type.option_widget is not None:
            self._options_widget = field_type.option_widget(self._opts_frame)
            self._options_widget.pack(fill="x")
            if options is not None and hasattr(self._options_widget, "set_value"):
                try:  # best effort
                    self._options_widget.set_value(options)  # type: ignore[attr-defined]
                except Exception:
                    pass
        elif field_type.option_model is not None:
            self._options_widget = OptionsForm(self._opts_frame, field_type)
            self._options_widget.pack(fill="x")
            if options is not None and hasattr(self._options_widget, "set_values"):
                try:  # best effort
                    self._options_widget.set_values(options)
                except Exception:
                    pass
        has_options = self._options_widget is not None and bool(self._opts_frame.winfo_children())
        if has_options:
            if self._opts_container is not None and not self._opts_container.winfo_manager():
                pack_opts = getattr(self, "_opts_pack", {"fill": "x", "pady": (0, 6)})
                self._opts_container.pack(**pack_opts)
        else:
            if self._opts_container is not None and self._opts_container.winfo_manager():
                self._opts_container.pack_forget()
        # Default editor
        if field_type.value_widget is not None:
            widget = field_type.value_widget(self._default_frame)  # type: ignore[assignment]
            widget.pack(fill="x")
            self._value_widget = widget
            if default is not None and hasattr(widget, "set_value"):
                try:  # best effort
                    widget.set_value(default)  # type: ignore[attr-defined]
                except Exception:
                    pass
        has_default = self._value_widget is not None and bool(self._default_frame.winfo_children())
        if has_default:
            if self._default_container is not None and not self._default_container.winfo_manager():
                pack_opts = getattr(self, "_default_pack", {"fill": "x", "pady": (0, 6)})
                self._default_container.pack(**pack_opts)
        else:
            if self._default_container is not None and self._default_container.winfo_manager():
                self._default_container.pack_forget()

    # ------------------------------------------------------------------
    def _on_type_change(self, _event: object | None = None) -> None:
        ft = TYPE_REGISTRY.get(self._type_var.get())
        default = None
        if self._value_widget is not None and hasattr(self._value_widget, "get_value"):
            try:
                default = self._value_widget.get_value()  # type: ignore[attr-defined]
            except Exception:
                default = None
        self._build_type_section(ft, default)
        self._register_form_watchers()
        self._update_dirty_state()

    def _load_tree_node(self, node: str) -> None:
        if ":" not in node:
            return
        _, key = node.split(":", 1)
        self._current_key = key
        defined = {f.key: f for f in self.adapter.list_defined()}
        info: FieldInfo | None = defined.get(key)
        default: object | None = None
        undiscovered = False
        if info is not None:
            default = self.adapter.default_for_key(key)
        else:
            und = {u.key: u for u in self.adapter.list_undiscovered()}
            uinfo = und.get(key)
            if uinfo is None:
                return
            info = FieldInfo(key=key, type=uinfo.guessed_type)
            default = uinfo.raw
            undiscovered = True
        self._populate_form(info, default, undiscovered=undiscovered)
        self._current_tree_selection = node

    # ------------------------------------------------------------------
    def _on_select(self, _event: object | None = None) -> None:
        if self._suspend_tree_select:
            return
        sel = self._tree.selection()
        if not sel:
            return
        node = sel[0]
        if node == self._current_tree_selection:
            return

        def proceed() -> None:
            self._select_tree_iid(node)
            self._load_tree_node(node)

        if self._dirty_any():
            self._restore_tree_selection()
            self._confirm_unsaved(proceed)
            return
        proceed()

    def _on_add(self) -> None:
        def perform() -> None:
            self._suspend_tree_select = True
            try:
                self._tree.selection_remove(self._tree.selection())
            finally:
                self._suspend_tree_select = False
            self._current_tree_selection = None
            self._current_key = ""
            info = FieldInfo(key="", type="string")
            self._populate_form(info, None, undiscovered=False)

        if self._dirty_any():
            self._confirm_unsaved(perform)
        else:
            perform()

    # ------------------------------------------------------------------
    def _create_section(
        self,
        title: str,
        *,
        pack: dict[str, Any] | None = None,
    ) -> tuple[tk.Widget, ttk.Frame, dict[str, Any]]:
        palette = get_palette()
        container = tk.Frame(self._form, bg=palette["bg"], highlightthickness=0, bd=0)
        pack_opts: dict[str, Any] = {"fill": "x", "pady": (0, 6)}
        if pack:
            pack_opts.update(pack)
        container.pack(**pack_opts)
        card = tk.Frame(
            container,
            bg=palette["card"],
            highlightthickness=1,
            highlightbackground=palette["card_edge"],
            highlightcolor=palette["card_edge"],
            bd=0,
        )
        card.pack(fill="both", expand=True)
        header = tk.Label(
            card,
            text=title,
            bg=palette["bg"],
            fg=palette["hdr_fg"],
            font=(None, 10, "bold"),
            anchor="w",
        )
        header.pack(fill="x", padx=12, pady=(8, 0))
        body = ttk.Frame(card, style="CardBody.TFrame", padding=(12, 8, 12, 12))
        body.pack(fill="both", expand=True)
        return container, body, pack_opts.copy()

    # ------------------------------------------------------------------
    def _populate_form(self, info: FieldInfo, default: object | None, *, undiscovered: bool) -> None:
        self._clear_form()
        palette = get_palette()

        # Identity -------------------------------------------------------
        _ident_container, ident, _ = self._create_section("Identity")
        self._key_var = tk.StringVar(value=info.key)
        ttk.Label(ident, text="Key: ", style="Card.TLabel").grid(row=0, column=0, sticky="w")
        ttk.Entry(ident, textvariable=self._key_var).grid(row=0, column=1, sticky="ew")
        self._label_var = tk.StringVar(value=info.label or "")
        ttk.Label(ident, text="Label: ", style="Card.TLabel").grid(row=1, column=0, sticky="w")
        ttk.Entry(ident, textvariable=self._label_var).grid(row=1, column=1, sticky="ew")
        ident.columnconfigure(1, weight=1)

        # Descriptions ---------------------------------------------------
        _desc_container, desc_fr, _ = self._create_section("Descriptions")
        self._desc_short_var = tk.StringVar(value=info.description_short or "")
        ttk.Label(desc_fr, text="Short: ", style="Card.TLabel").grid(row=0, column=0, sticky="w")
        short_entry = ttk.Entry(desc_fr, textvariable=self._desc_short_var)
        short_entry.grid(row=0, column=1, sticky="ew")
        self._desc_short_count = ttk.Label(desc_fr, text="0/0", style="CardMuted.TLabel")
        self._desc_short_count.grid(row=0, column=2, sticky="e")
        desc_fr.columnconfigure(1, weight=1)

        def _update_count(*_args: object) -> None:
            n = len(self._desc_short_var.get())
            fg = "red" if n > SHORT_DESC_MAX else get_palette()['hdr_fg']
            self._desc_short_count.config(text=f"{n}/{SHORT_DESC_MAX}", foreground=fg)

        self._desc_short_var.trace_add("write", _update_count)
        _update_count()

        ttk.Label(desc_fr, text="Long: ", style="Card.TLabel").grid(row=1, column=0, sticky="nw")
        self._desc_text = tk.Text(
            desc_fr,
            height=4,
            wrap="word",
            bg=palette["field"],
            fg=palette["ink"],
            insertbackground=palette["ink"],
            highlightthickness=1,
            highlightbackground=palette["field_bd"],
            highlightcolor=palette["primary"],
            relief="flat",
            bd=1,
        )
        self._desc_text.grid(row=1, column=1, columnspan=2, sticky="ew")
        if info.description:
            self._desc_text.insert("1.0", info.description)

        # Type -----------------------------------------------------------
        _type_container, type_fr, _ = self._create_section("Type")
        self._type_var = tk.StringVar(value=info.type)
        ttk.Label(type_fr, text="Type: ", style="Card.TLabel").grid(row=0, column=0, sticky="w")
        type_combo = ttk.Combobox(
            type_fr,
            textvariable=self._type_var,
            state="readonly",
            values=sorted(TYPE_REGISTRY.keys()),
        )
        type_combo.grid(row=0, column=1, sticky="ew")
        type_combo.bind("<<ComboboxSelected>>", self._on_type_change)
        type_fr.columnconfigure(1, weight=1)

        opts_container, self._opts_frame, opts_pack = self._create_section("Type Options")
        self._opts_container = opts_container
        self._opts_pack = opts_pack
        default_container, self._default_frame, default_pack = self._create_section("Default")
        self._default_container = default_container
        self._default_pack = default_pack
        ft = TYPE_REGISTRY.get(info.type)
        self._build_type_section(ft, default, info.options)

        # Grouping ------------------------------------------------------
        _group_container, group_fr, _ = self._create_section("Grouping")
        self._section_var = tk.StringVar(value=info.section or "")
        ttk.Label(group_fr, text="Section: ", style="Card.TLabel").grid(row=0, column=0, sticky="w")
        ttk.Entry(group_fr, textvariable=self._section_var).grid(row=0, column=1, sticky="ew")
        self._order_var = tk.StringVar(
            value="" if info.order is None else str(info.order)
        )
        ttk.Label(group_fr, text="Order: ", style="Card.TLabel").grid(row=1, column=0, sticky="w")
        ttk.Entry(group_fr, textvariable=self._order_var).grid(row=1, column=1, sticky="ew")
        group_fr.columnconfigure(1, weight=1)

        # Actions --------------------------------------------------------
        actions = ttk.Frame(self._form, style="TFrame")
        actions.pack(fill="x", pady=(0, 6))
        ttk.Button(actions, text="Save", command=self._on_save).pack(side="right")
        ttk.Button(actions, text="Revert", command=self._on_revert).pack(side="right")
        ttk.Button(actions, text="Delete", command=self._on_delete).pack(side="right")
        if undiscovered:
            ttk.Button(actions, text="Adopt", command=self._on_adopt).pack(side="right")

        # Diff drawer ----------------------------------------------------
        self._diff_shown = tk.BooleanVar(value=False)
        toggle = ttk.Checkbutton(
            self._form,
            text="Show Diff",
            variable=self._diff_shown,
            command=self._toggle_diff,
            style="Toolbutton",
        )
        toggle.pack(anchor="w")
        self._diff_frame = ttk.Frame(self._form, style="TFrame")
        self._diff_text = tk.Text(
            self._diff_frame,
            height=6,
            bg=palette["field"],
            fg=palette["ink"],
            insertbackground=palette["ink"],
            highlightthickness=1,
            highlightbackground=palette["field_bd"],
            highlightcolor=palette["primary"],
            relief="flat",
            bd=1,
        )
        self._diff_text.pack(fill="both", expand=True)
        self._register_form_watchers()
        self._store_field_snapshot()

    # ------------------------------------------------------------------
    # Action handlers
    # ------------------------------------------------------------------
    def _collect_options(self) -> dict | None:
        ow = self._options_widget
        if ow is None:
            return None
        if hasattr(ow, "get_values"):
            return ow.get_values()  # type: ignore[no-any-return]
        if hasattr(ow, "get_value"):
            val = ow.get_value()  # type: ignore[no-untyped-call]
            return val if isinstance(val, dict) else {}
        return None

    def _try_save_field(self) -> bool:
        if self._current_key is None:
            return False
        key = self._key_var.get().strip()
        type_name = self._type_var.get().strip()
        label = self._label_var.get().strip() or None
        desc_short = self._desc_short_var.get().strip() or None
        description = self._desc_text.get("1.0", "end").strip() or None
        options = self._collect_options()
        section = self._section_var.get().strip() or None
        order_val = self._order_var.get().strip()
        order = None
        if order_val:
            try:
                order = int(order_val)
            except ValueError:
                if messagebox is not None:
                    messagebox.showerror("Invalid order", "Order must be an integer", parent=self)
                return False
        default = None
        if self._value_widget is not None and hasattr(self._value_widget, "get_value"):
            default = self._value_widget.get_value()  # type: ignore[attr-defined]
        if default is not None:
            try:
                default = parse_field_value(type_name, default)
            except (TypeError, ValueError):
                if messagebox is not None:
                    if type_name == "boolean":
                        msg = "Default must be true/false or 1/0"
                    else:
                        msg = f"Default must be a {type_name}"
                    messagebox.showerror("Invalid default", msg, parent=self)
                return False
        import inspect

        sig = inspect.signature(self.adapter.upsert_field)
        kwargs: dict[str, object] = {}
        if "label" in sig.parameters:
            kwargs["label"] = label
        if "description_short" in sig.parameters:
            kwargs["description_short"] = desc_short
        if "description" in sig.parameters:
            kwargs["description"] = description
        if "options" in sig.parameters and options is not None:
            kwargs["options"] = options
        if "section" in sig.parameters:
            kwargs["section"] = section
        if "order" in sig.parameters and order is not None:
            kwargs["order"] = order
        if "default" in sig.parameters and default is not None:
            kwargs["default"] = default
        if desc_short is not None and len(desc_short) > SHORT_DESC_MAX:
            if messagebox is not None:
                messagebox.showerror(
                    "Too long", f"Short description exceeds {SHORT_DESC_MAX} characters", parent=self
                )
            return False
        self.adapter.upsert_field(key, type_name, **kwargs)  # type: ignore[arg-type]
        self._current_key = key
        target = f"defined:{key}" if key else None
        if target is not None:
            self._current_tree_selection = target
        self._reload_tree()
        defined = {f.key: f for f in self.adapter.list_defined()}
        info = defined.get(key)
        default_val = self.adapter.default_for_key(key)
        if info is not None:
            self._populate_form(info, default_val, undiscovered=False)
        if target is not None:
            self._select_tree_iid(target)
        return True

    def _on_save(self) -> None:
        self._try_save_field()

    def _on_revert(self) -> None:
        if self._current_key is not None:
            defined = {f.key: f for f in self.adapter.list_defined()}
            info = defined.get(self._current_key)
            default = self.adapter.default_for_key(self._current_key)
            if info is not None:
                self._populate_form(info, default, undiscovered=False)

    def _on_delete(self) -> None:
        if self._current_key:
            try:
                self.adapter.delete_field(self._current_key)
            except Exception:
                pass
            self._current_key = None
            self._current_tree_selection = None
            self._reload_tree()
        self._clear_form()

    def _on_adopt(self) -> None:
        if self._current_key is None:
            return
        key = self._key_var.get().strip()
        type_name = self._type_var.get().strip()
        try:
            self.adapter.adopt_untracked({key: type_name})
        except Exception:
            return
        self._current_key = key
        target = f"defined:{key}" if key else None
        if target is not None:
            self._current_tree_selection = target
        self._reload_tree()
        defined = {f.key: f for f in self.adapter.list_defined()}
        info = defined.get(key)
        default = self.adapter.default_for_key(key)
        if info is not None:
            self._populate_form(info, default, undiscovered=False)
        if target is not None:
            self._select_tree_iid(target)

    def _toggle_diff(self) -> None:
        if self._diff_shown.get():
            self._diff_frame.pack(fill="both", expand=True)
        else:
            self._diff_frame.pack_forget()



import pytest
from pathlib import PurePosixPath
from types import SimpleNamespace
import pysigil.settings_metadata as settings_metadata
from pysigil.errors import ConflictError
from pysigil.settings_metadata import (
    FieldSpec,
    IniFileBackend,
    IniSpecBackend,
    ProviderManager,
    ProviderSpec,
)
from pysigil.io_config import read_sections, write_sections
from tests.utils import DummyPolicy


def test_fieldspec_validates_unknown_type():
    with pytest.raises(ValueError):
        FieldSpec(key="foo", type="unknown")


class DummyBackend:
    def __init__(self):
        self.writes = []
        self.removes = []
        self.sections = []

    def read_merged(self, provider_id):
        return {"retries": "3"}, {"retries": "project"}

    def write_key(self, provider_id, key, raw_value, *, scope, target_kind):
        self.writes.append((provider_id, key, raw_value, scope, target_kind))

    def remove_key(self, provider_id, key, *, scope, target_kind):
        self.removes.append((provider_id, key, scope, target_kind))

    def ensure_section(self, provider_id, *, scope, target_kind):
        self.sections.append((provider_id, scope, target_kind))

    def write_target_for(self, provider_id):
        return "settings.ini"


def test_provider_manager_roundtrip():
    spec = ProviderSpec(
        provider_id="demo",
        schema_version="0.1",
        fields=[
            FieldSpec(
                key="retries",
                type="integer",
                label="Retries",
                options={"minimum": 0},
            )
        ],
    )
    backend = DummyBackend()
    mgr = ProviderManager(spec, backend)

    state = mgr.effective()
    assert state["retries"].value == 3
    assert state["retries"].source == "project"
    assert mgr.spec.fields[0].options == {"minimum": 0}

    mgr.set("retries", 5)
    assert backend.writes == [("demo", "retries", "5", "user", "settings.ini")]

    mgr.clear("retries")
    assert backend.removes == [("demo", "retries", "user", "settings.ini")]

    mgr.init("user")
    assert backend.sections == [("demo", "user", "settings.ini")]


def test_integer_minimum_enforced():
    spec = ProviderSpec(
        provider_id="demo",
        schema_version="0.1",
        fields=[FieldSpec(key="count", type="integer", options={"minimum": 1})],
    )
    backend = DummyBackend()
    mgr = ProviderManager(spec, backend)
    with pytest.raises(ValueError):
        mgr.set("count", 0)


def test_ini_spec_backend_persists_options(tmp_path):
    backend = IniSpecBackend(user_dir=tmp_path)
    spec = ProviderSpec(
        provider_id="demo",
        schema_version="1",
        fields=[
            FieldSpec(key="mode", type="string", options={"choices": ["a", "b"]})
        ],
    )
    backend.create_spec(spec)
    loaded = backend.get_spec("demo")
    assert loaded.fields[0].options == {"choices": ["a", "b"]}


def test_ini_spec_backend_persists_sections(tmp_path):
    backend = IniSpecBackend(user_dir=tmp_path)
    spec = ProviderSpec(
        provider_id="demo",
        schema_version="1",
        sections_order=["Basics", "Advanced"],
        sections_collapsed=["Advanced"],
        fields=[
            FieldSpec(
                key="endpoint", type="string", section="Basics", order=100
            )
        ],
    )
    backend.create_spec(spec)
    loaded = backend.get_spec("demo")
    assert loaded.sections_order == ("Basics", "Advanced")
    assert loaded.sections_collapsed == ("Advanced",)
    assert loaded.fields[0].section == "Basics"
    assert loaded.fields[0].order == 100


def test_ini_file_backend(tmp_path):
    user_dir = tmp_path / "user"
    project_dir = tmp_path / "proj"
    policy = DummyPolicy(user_dir, project_dir, host="host")
    backend = IniFileBackend(policy=policy)

    write_sections(user_dir / "demo" / "settings.ini", {"demo": {"alpha": "u"}})
    write_sections(project_dir / "settings.ini", {"demo": {"alpha": "p", "beta": "p"}})
    write_sections(
        project_dir / "settings-local-host.ini", {"demo": {"beta": "pl"}}
    )

    raw, src = backend.read_merged("demo")
    assert raw == {"alpha": "p", "beta": "pl"}
    assert src == {"alpha": "project", "beta": "project-local"}

    backend.ensure_section("demo", scope="user", target_kind="settings.ini")
    backend.write_key("demo", "gamma", "42", scope="user", target_kind="settings.ini")
    data = read_sections(user_dir / "demo" / "settings.ini")
    assert data["demo"]["gamma"] == "42"

    backend.remove_key("demo", "gamma", scope="user", target_kind="settings.ini")
    data = read_sections(user_dir / "demo" / "settings.ini")
    assert "gamma" not in data["demo"]

    assert backend.write_target_for("user-custom") == "settings-local-host.ini"
    assert backend.write_target_for("demo") == "settings.ini"


def test_ini_file_backend_prefers_installed_defaults(monkeypatch, tmp_path):
    user_dir = tmp_path / "user"
    project_dir = tmp_path / "proj"
    policy = DummyPolicy(user_dir, project_dir, host="host")
    backend = IniFileBackend(policy=policy)

    dev_defaults = tmp_path / "dev" / ".sigil" / "settings.ini"
    dev_defaults.parent.mkdir(parents=True, exist_ok=True)
    write_sections(dev_defaults, {"demo": {"alpha": "dev"}})

    installed_defaults = tmp_path / "installed" / ".sigil" / "settings.ini"
    installed_defaults.parent.mkdir(parents=True, exist_ok=True)
    write_sections(installed_defaults, {"demo": {"alpha": "installed"}})

    monkeypatch.setattr(
        settings_metadata,
        "get_dev_link",
        lambda pid: SimpleNamespace(defaults_path=dev_defaults),
    )

    def fake_resolve(provider_id, filename="settings.ini"):
        assert filename == "settings.ini"
        return installed_defaults, "installed"

    monkeypatch.setattr(settings_metadata, "resolve_defaults", fake_resolve)

    layers = backend.read_layers("demo")
    assert layers["default"]["alpha"] == "installed"


def test_short_description_length_limit():
    with pytest.raises(ValueError):
        FieldSpec(key="k", type="string", description_short="x" * 121)


def test_fieldspec_serialization_omits_empty_descriptions():
    spec = FieldSpec(key="k", type="string")
    gui = spec.to_gui_v0()
    assert "description" not in gui
    assert "description_short" not in gui
    spec = FieldSpec(key="k", type="string", description="long")
    gui = spec.to_gui_v0()
    assert gui["description"] == "long"
    assert "description_short" not in gui


def test_spec_backend_detects_external_change(tmp_path):
    backend = IniSpecBackend(user_dir=tmp_path)
    spec = ProviderSpec(provider_id="demo", schema_version="1")
    backend.create_spec(spec)
    loaded = backend.get_spec("demo")
    path = tmp_path / "demo" / "metadata.ini"
    path.write_text(path.read_text() + "\n# external\n")
    with pytest.raises(ConflictError):
        backend.save_spec(loaded)


def test_ini_spec_backend_get_provider_ids_installed(monkeypatch):
    backend = IniSpecBackend()
    monkeypatch.setattr(settings_metadata, "list_links", lambda **kw: {})

    class DummyDist:
        def __init__(self, name, files, meta_name=None):
            self.name = name
            self._files = [PurePosixPath(p) for p in files]
            if meta_name is None:
                self.metadata = {}
            else:
                self.metadata = {"Name": meta_name}

        @property
        def files(self):  # pragma: no cover - simple
            return self._files

    dists = [
        DummyDist("pkg_a", [".sigil/metadata.ini"], "Foo"),
        DummyDist("pkg_b", ["other.txt"], "Bar"),
    ]
    monkeypatch.setattr(settings_metadata, "distributions", lambda: dists)

    ids = backend.get_provider_ids()
    assert ids == ["foo"]


def test_ini_spec_backend_prefers_dev_link(monkeypatch, tmp_path):
    defaults = tmp_path / "foo" / "defaults.ini"
    defaults.parent.mkdir(parents=True)
    defaults.write_text("")
    (defaults.parent / "metadata.ini").write_text("[__meta__]\nschema_version=1\n")

    monkeypatch.setattr(settings_metadata, "list_links", lambda **kw: {"foo": defaults})

    class DummyDist:
        def __init__(self, name):
            self.name = name
            self._files = [PurePosixPath(".sigil/metadata.ini")]
            self.metadata = {"Name": name}

        @property
        def files(self):  # pragma: no cover - simple
            return self._files

    dists = [DummyDist("Foo"), DummyDist("Bar")]
    monkeypatch.setattr(settings_metadata, "distributions", lambda: dists)

    backend = IniSpecBackend()
    ids = backend.get_provider_ids()
    assert ids == ["bar", "foo"]





"""Widget registry for basic field editors.

The registry maps a field ``type`` as reported by :mod:`pysigil.api` to a
callable returning an editor widget.  Editors follow a very small
protocol which makes it straightforward to provide alternative
implementations for other GUI toolkits.
"""

from __future__ import annotations

from dataclasses import replace
from typing import Callable, Protocol, Dict, Any

from ..settings_metadata import TYPE_REGISTRY

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk
    from tkinter import ttk
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    ttk = None  # type: ignore


class EditorWidget(Protocol):
    """Protocol all editor widgets must implement."""

    def get_value(self) -> object | None: ...
    def set_value(self, value: object | None) -> None: ...
    def set_error(self, msg: str | None) -> None: ...
    def set_source_badge(self, source: str | None) -> None: ...


# -- concrete editor implementations ---------------------------------------


def _simple_entry(master) -> EditorWidget:
    if tk is None or ttk is None:  # pragma: no cover - tkinter missing
        raise RuntimeError("tkinter is required for widgets")
    frame = ttk.Frame(master, style="CardBody.TFrame")
    entry = ttk.Entry(frame)
    entry.pack(side="left", fill="x", expand=True)
    badge = ttk.Label(frame, text="", style="CardMuted.TLabel")
    badge.pack(side="left", padx=4)

    def get_value() -> object | None:
        text = entry.get()
        return text if text != "" else None

    def set_value(value: object | None) -> None:
        entry.delete(0, tk.END)
        if value is not None:
            entry.insert(0, str(value))

    def set_error(msg: str | None) -> None:
        if msg:
            entry.configure(foreground="red")
        else:
            entry.configure(foreground="black")

    def set_source_badge(src: str | None) -> None:
        badge.configure(text=src or "")

    frame.get_value = get_value  # type: ignore[attr-defined]
    frame.set_value = set_value  # type: ignore[attr-defined]
    frame.set_error = set_error  # type: ignore[attr-defined]
    frame.set_source_badge = set_source_badge  # type: ignore[attr-defined]
    return frame  # type: ignore[return-value]


def _boolean_tristate(master) -> EditorWidget:
    if tk is None or ttk is None:  # pragma: no cover - tkinter missing
        raise RuntimeError("tkinter is required for widgets")
    frame = ttk.Frame(master, style="CardBody.TFrame")
    var = tk.StringVar(value="unset")
    btns = ttk.Frame(frame, style="CardBody.TFrame")
    btns.pack(side="left")
    ttk.Radiobutton(btns, text="True", variable=var, value="true", style="Card.TRadiobutton").pack(side="left")
    ttk.Radiobutton(btns, text="False", variable=var, value="false", style="Card.TRadiobutton").pack(side="left")
    ttk.Radiobutton(btns, text="Unset", variable=var, value="unset", style="Card.TRadiobutton").pack(side="left")
    badge = ttk.Label(frame, text="", style="CardMuted.TLabel")
    badge.pack(side="left", padx=4)

    def get_value() -> object | None:
        val = var.get()
        if val == "unset":
            return None
        return val == "true"

    def set_value(value: object | None) -> None:
        if value is True:
            var.set("true")
        elif value is False:
            var.set("false")
        else:
            var.set("unset")

    def set_error(_msg: str | None) -> None:  # pragma: no cover - placeholder
        pass

    def set_source_badge(src: str | None) -> None:
        badge.configure(text=src or "")

    frame.get_value = get_value  # type: ignore[attr-defined]
    frame.set_value = set_value  # type: ignore[attr-defined]
    frame.set_error = set_error  # type: ignore[attr-defined]
    frame.set_source_badge = set_source_badge  # type: ignore[attr-defined]
    return frame  # type: ignore[return-value]


# Register widget implementations with the type registry
TYPE_REGISTRY["string"] = replace(TYPE_REGISTRY["string"], value_widget=_simple_entry)
TYPE_REGISTRY["integer"] = replace(TYPE_REGISTRY["integer"], value_widget=_simple_entry)
TYPE_REGISTRY["number"] = replace(TYPE_REGISTRY["number"], value_widget=_simple_entry)
TYPE_REGISTRY["boolean"] = replace(TYPE_REGISTRY["boolean"], value_widget=_boolean_tristate)
TYPE_REGISTRY["string_list"] = replace(TYPE_REGISTRY["string_list"], value_widget=_simple_entry)


FIELD_WIDGETS: Dict[str, Callable[[Any], EditorWidget]] = {
    key: ft.value_widget  # type: ignore[assignment]
    for key, ft in TYPE_REGISTRY.items()
    if ft.value_widget is not None
}


__all__ = ["EditorWidget", "FIELD_WIDGETS"]



import os
from pathlib import Path

import pytest

from pysigil import Sigil
from pysigil.policy import policy
from pysigil.errors import ReadOnlyScopeError


def test_roundtrip(tmp_path: Path) -> None:
    s = Sigil("demo", user_scope=tmp_path, policy=policy)
    s.set_pref("foo.bar", "baz", scope="user")
    assert s.get_pref("foo.bar") == "baz"
    content = (tmp_path / "settings.ini").read_text()
    assert "[demo]" in content
    assert "foo_bar = baz" in content
def test_package_defaults_read_only(tmp_path: Path, monkeypatch) -> None:
    pkg = tmp_path / "pkgdefaults"
    (pkg / ".sigil").mkdir(parents=True)
    (pkg / ".sigil" / "settings.ini").write_text("[pkgdefaults]\nfoo = 7\n")
    (pkg / "__init__.py").write_text("")
    monkeypatch.syspath_prepend(tmp_path)
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
    user_dir = tmp_path / "user"
    s = Sigil("pkgdefaults", user_scope=user_dir, policy=policy)
    # assert s.get_pref("foo") == 7
    with pytest.raises(ReadOnlyScopeError):
        s.set_pref("foo", "8", scope="default")


def test_dev_link_defaults_writable(tmp_path: Path, monkeypatch) -> None:
    pkg = tmp_path / "pkgdefaults"
    (pkg / ".sigil").mkdir(parents=True)
    settings = pkg / ".sigil" / "settings.ini"
    settings.write_text("[pkgdefaults]\nfoo = 7\n")
    (pkg / "__init__.py").write_text("")
    monkeypatch.syspath_prepend(tmp_path)
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
    from pysigil.authoring import link

    link("pkgdefaults", settings)
    user_dir = tmp_path / "user"
    s = Sigil("pkgdefaults", user_scope=user_dir, policy=policy)
    s.set_pref("foo", "8", scope="default")
    assert s.get_pref("foo") == 8
    assert "foo = 8" in settings.read_text()


def test_explicit_default_path_writable(tmp_path: Path) -> None:
    default_dir = tmp_path / "defaults"
    s = Sigil("demo", user_scope=tmp_path / "user.ini", default_path=default_dir, policy=policy)
    s.set_pref("foo", "bar", scope="default")
    assert s.get_pref("foo") == "bar"

    settings_file = default_dir / ".sigil" / "settings.ini"

    assert "foo = bar" in settings_file.read_text()
    s2 = Sigil("demo", user_scope=tmp_path / "user2.ini", default_path=default_dir, policy=policy)
    assert s2.get_pref("foo") == "bar"


def test_environment_scope_updates_environ(monkeypatch, tmp_path: Path) -> None:
    sigil = Sigil("demo", user_scope=tmp_path / "user", policy=policy)
    env_key = "SIGIL_DEMO_SECTION_VALUE"
    monkeypatch.delenv(env_key, raising=False)

    try:
        sigil.set_pref("section.value", "123", scope="env")
        assert os.environ[env_key] == "123"
        assert sigil.get_pref("section.value") == 123
    finally:
        sigil.set_pref("section.value", None, scope="env")
    assert env_key not in os.environ
    assert sigil.get_pref("section.value") is None




from pysigil import authoring
from pysigil.ui.author_adapter import AuthorAdapter
import pysigil.api as api


def test_upsert_field_sets_options_and_default(tmp_path, monkeypatch):
    monkeypatch.setenv("SIGIL_APP_NAME", "sigil-test")
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    monkeypatch.setenv("SIGIL_ROOT", str(tmp_path))

    defaults_dir = tmp_path / "pkg" / ".sigil"
    defaults_dir.mkdir(parents=True)
    defaults_path = defaults_dir / "settings.ini"
    defaults_path.write_text("[demo]\n")

    authoring.link("demo", defaults_path)
    api.register_provider("demo", title="Demo")

    adapter = AuthorAdapter("demo")
    adapter.upsert_field("alpha", "integer", options={"minimum": 0}, default=5)

    fields = {f.key: f for f in adapter.list_defined()}
    assert fields["alpha"].options == {"minimum": 0}
    assert adapter.default_for_key("alpha") == 5


def test_adopt_untracked(tmp_path, monkeypatch):
    monkeypatch.setenv("SIGIL_APP_NAME", "sigil-test")
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    monkeypatch.setenv("SIGIL_ROOT", str(tmp_path))

    defaults_dir = tmp_path / "pkg" / ".sigil"
    defaults_dir.mkdir(parents=True)
    defaults_path = defaults_dir / "settings.ini"
    defaults_path.write_text("[demo]\nfoo = bar\n")

    authoring.link("demo", defaults_path)
    api.register_provider("demo", title="Demo")

    adapter = AuthorAdapter("demo")
    assert any(u.key == "foo" for u in adapter.list_undiscovered())

    adapter.adopt_untracked({"foo": "string"})
    defined = {f.key for f in adapter.list_defined()}
    assert "foo" in defined
    assert not any(u.key == "foo" for u in adapter.list_undiscovered())




from __future__ import annotations

import argparse
import tomllib
from pathlib import Path

import pysigil.authoring as authoring
from pysigil import cli
from pysigil.authoring import ensure_sigil_package_data


def _write_pyproject(root: Path, name: str) -> None:
    content = f"[project]\nname = \"{name}\"\n"
    (root / "pyproject.toml").write_text(content, encoding="utf-8")


def test_ensure_sigil_package_data(tmp_path: Path) -> None:
    _write_pyproject(tmp_path, "demo")
    ensure_sigil_package_data(tmp_path, "demo_pkg")
    data = tomllib.loads((tmp_path / "pyproject.toml").read_text(encoding="utf-8"))
    assert data["tool"]["setuptools"]["package-data"]["demo_pkg"] == [".sigil/*"]
    # idempotent
    ensure_sigil_package_data(tmp_path, "demo_pkg")
    data = tomllib.loads((tmp_path / "pyproject.toml").read_text(encoding="utf-8"))
    assert data["tool"]["setuptools"]["package-data"]["demo_pkg"] == [".sigil/*"]


def test_author_register_adds_package_data(monkeypatch, tmp_path: Path) -> None:
    root = tmp_path
    _write_pyproject(root, "demo-pkg")
    pkg = root / "src" / "demo_pkg"
    pkg.mkdir(parents=True)
    (pkg / "__init__.py").touch()

    monkeypatch.setattr(authoring, "_dev_dir", lambda: tmp_path / "dev")
    monkeypatch.chdir(root)

    ns = argparse.Namespace(
        auto=True,
        package_dir=None,
        defaults=None,
        provider=None,
        no_validate=True,
    )
    assert cli.author_register(ns) == 0
    data = tomllib.loads((root / "pyproject.toml").read_text(encoding="utf-8"))
    assert data["tool"]["setuptools"]["package-data"]["demo_pkg"] == [".sigil/*"]




"""Automated tests for the :mod:`pysigil.orchestrator` module."""

from pathlib import Path

import pytest

from pysigil.orchestrator import Orchestrator, PolicyError, ValidationError
from pysigil.settings_metadata import IniFileBackend, IniSpecBackend, read_sections, write_sections
import pysigil.settings_metadata as sm
from tests.utils import DummyPolicy


def _make_orch(tmp_path: Path) -> Orchestrator:
    spec_backend = IniSpecBackend(user_dir=tmp_path / "user")
    policy = DummyPolicy(tmp_path / "user", tmp_path / "proj", host="host")
    cfg_backend = IniFileBackend(policy=policy)
    return Orchestrator(spec_backend=spec_backend, config_backend=cfg_backend)


def test_register_add_set_get(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("my-pkg", title="My Package")
    orch.add_field(
        "my-pkg",
        key="retries",
        type="integer",
        label="Retries",
        options={"minimum": 0},
    )
    orch.set_value("my-pkg", "retries", 5)
    eff = orch.get_effective("my-pkg")
    assert eff["retries"].value == 5
    assert eff["retries"].source == "user"
    spec = orch.reload_spec("my-pkg")
    assert spec.fields[0].options == {"minimum": 0}


def test_set_value_respects_minimum(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field(
        "pkg",
        key="retries",
        type="integer",
        options={"minimum": 1},
    )
    with pytest.raises(ValidationError):
        orch.set_value("pkg", "retries", 0)


def test_edit_field_rename_migrates_value(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="alpha", type="string")
    orch.set_value("pkg", "alpha", "one")
    orch.edit_field("pkg", "alpha", new_key="beta")
    eff = orch.get_effective("pkg")
    assert "beta" in eff and eff["beta"].value == "one"
    path = tmp_path / "user" / "pkg" / "settings.ini"
    assert path.read_text().strip().endswith("beta = one")


def test_edit_field_type_change_convert(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="num", type="string")
    orch.set_value("pkg", "num", "42")
    orch.edit_field("pkg", "num", new_type="integer", on_type_change="convert")
    eff = orch.get_effective("pkg")
    assert eff["num"].value == 42


def test_edit_field_type_change_convert_failure(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="num", type="string")
    orch.set_value("pkg", "num", "forty-two")
    with pytest.raises(ValidationError):
        orch.edit_field("pkg", "num", new_type="integer", on_type_change="convert")
    spec = orch.reload_spec("pkg")
    assert spec.fields[0].type == "string"
    eff = orch.get_effective("pkg")
    assert eff["num"].value == "forty-two"


def test_delete_field_removes_values(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="alpha", type="string")
    orch.set_value("pkg", "alpha", "x")
    orch.delete_field("pkg", "alpha", remove_values=True)
    assert orch.list_fields("pkg") == []
    path = tmp_path / "user" / "pkg" / "settings.ini"
    assert "alpha" not in path.read_text()


def test_find_and_adopt_untracked(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    target = orch.config_backend.write_target_for("pkg")
    orch.config_backend.write_key(
        "pkg", "foo", "bar", scope="user", target_kind=target
    )
    assert orch.find_untracked_keys("pkg") == ["foo"]
    orch.adopt_untracked("pkg", {"foo": "string"})
    eff = orch.get_effective("pkg")
    assert eff["foo"].value == "bar"


def test_validate_all_reports_errors(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="num", type="integer")
    target = orch.config_backend.write_target_for("pkg")
    orch.config_backend.write_key(
        "pkg", "num", "oops", scope="user", target_kind=target
    )
    errors = orch.validate_all("pkg")
    assert errors["num"] is not None


def test_patch_fields_and_sections(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="a", type="string", section="Old", order=1)
    orch.add_field("pkg", key="b", type="string", section="Old", order=2)
    orch.patch_fields("pkg", [{"key": "b", "section": "New", "order": 5}])
    spec = orch.reload_spec("pkg")
    b_field = [f for f in spec.fields if f.key == "b"][0]
    assert b_field.section == "New" and b_field.order == 5
    orch.set_sections_order("pkg", ["New", "Old"])
    assert orch.get_sections_order("pkg") == ["New", "Old"]
    orch.set_sections_collapsed("pkg", ["Old"])
    assert orch.get_sections_collapsed("pkg") == ["Old"]


def test_set_many_atomic(tmp_path: Path) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="a", type="integer")
    orch.add_field("pkg", key="b", type="integer")
    with pytest.raises(ValidationError):
        orch.set_many("pkg", {"a": 1, "b": "bad"}, atomic=True)
    eff = orch.get_effective("pkg")
    assert eff["a"].value is None and eff["b"].value is None


def test_set_many_rollback_on_write_error(tmp_path: Path, monkeypatch) -> None:
    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="a", type="integer")
    orch.add_field("pkg", key="b", type="integer")
    target = tmp_path / "user" / "pkg" / "settings.ini"
    write_sections(target, {"pkg": {"a": "1"}})

    def boom(path, data):
        raise RuntimeError("boom")

    monkeypatch.setattr(sm, "write_sections", boom)
    with pytest.raises(RuntimeError):
        orch.set_many("pkg", {"a": 2, "b": 3}, atomic=True)
    assert read_sections(target) == {"pkg": {"a": "1"}}


def test_default_scope_editing_with_dev_link(tmp_path: Path, monkeypatch) -> None:
    import pysigil.authoring as auth

    user_dir = tmp_path / "user"
    dev_defaults = tmp_path / "pkg" / ".sigil" / "settings.ini"
    dev_defaults.parent.mkdir(parents=True)
    dev_defaults.write_text("[pkg]\nkey=0\n")
    monkeypatch.setattr(auth, "user_config_dir", lambda app: str(user_dir))
    auth.link("pkg", dev_defaults, validate=False)

    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="key", type="integer")
    orch.set_value("pkg", "key", 5, scope="default")
    eff = orch.get_effective("pkg")
    assert eff["key"].source == "default"
    assert eff["key"].value == 5
    assert "key = 5" in dev_defaults.read_text()


def test_metadata_requires_dev_link(tmp_path: Path) -> None:
    policy = DummyPolicy(tmp_path / "user", tmp_path / "proj", host="host")
    cfg_backend = IniFileBackend(policy=policy)
    orch = Orchestrator(config_backend=cfg_backend)
    with pytest.raises(PolicyError):
        orch.register_provider("pkg")


def test_metadata_stored_in_defaults(tmp_path: Path, monkeypatch) -> None:
    import pysigil.authoring as auth

    user_dir = tmp_path / "user"
    dev_defaults = tmp_path / "pkg" / ".sigil" / "settings.ini"
    dev_defaults.parent.mkdir(parents=True)
    dev_defaults.write_text("[pkg]\n")
    monkeypatch.setattr(auth, "user_config_dir", lambda app: str(user_dir))
    auth.link("pkg", dev_defaults, validate=False)

    policy = DummyPolicy(user_dir, tmp_path / "proj", host="host")
    cfg_backend = IniFileBackend(policy=policy)
    orch = Orchestrator(config_backend=cfg_backend)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="alpha", type="string")
    meta_path = dev_defaults.parent / "metadata.ini"
    assert meta_path.exists()
    assert "field:alpha" in meta_path.read_text()


def test_scope_precedence(tmp_path: Path, monkeypatch) -> None:
    import pysigil.authoring as auth

    user_dir = tmp_path / "user"
    dev_defaults = tmp_path / "pkg" / ".sigil" / "settings.ini"
    dev_defaults.parent.mkdir(parents=True)
    dev_defaults.write_text("[pkg]\n", encoding="utf-8")
    monkeypatch.setattr(auth, "user_config_dir", lambda app: str(user_dir))
    auth.link("pkg", dev_defaults, validate=False)

    orch = _make_orch(tmp_path)
    orch.register_provider("pkg")
    orch.add_field("pkg", key="key", type="string")

    orch.set_value("pkg", "key", "default", scope="default")
    orch.set_value("pkg", "key", "user", scope="user")
    orch.set_value("pkg", "key", "project", scope="project")
    eff = orch.get_effective("pkg")
    assert eff["key"].value == "project"
    assert eff["key"].source == "project"

    orch.clear_value("pkg", "key", scope="project")

    eff = orch.get_effective("pkg")
    assert eff["key"].value == "user"
    assert eff["key"].source == "user"

    orch.clear_value("pkg", "key", scope="user")
    eff = orch.get_effective("pkg")

    assert eff["key"].value == "default"
    assert eff["key"].source == "default"




import pytest

from pysigil import Sigil
from pysigil.authoring import normalize_provider_id
from pysigil.discovery import pep503_name
from pysigil.policy import policy


@pytest.mark.parametrize("name", ["foo/bar", "foo\\bar", "foo..bar"])
def test_pep503_name_rejects_invalid(name):
    with pytest.raises(ValueError):
        pep503_name(name)


@pytest.mark.parametrize("name", ["foo/bar", "foo\\bar", "foo..bar"])
def test_normalize_provider_id_rejects_invalid(name):
    with pytest.raises(ValueError):
        normalize_provider_id(name)


@pytest.mark.parametrize("name", ["foo/bar", "foo\\bar", "foo..bar"])
def test_sigil_rejects_invalid_app_name(name, tmp_path):
    with pytest.raises(ValueError):
        Sigil(name, user_scope=tmp_path / "user.ini", policy=policy)



"""User interface helpers for pysigil.

This package contains a framework agnostic core layer and optional
front-end implementations.  The initial implementation uses tkinter
but the design allows the view layer to be swapped out with minimal
changes to the core logic.
"""

from __future__ import annotations

__all__ = ["core", "tk", "widgets", "options_form"]



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

import pysigil.paths as _paths

from .core import Sigil

__all__ = ["helpers_for", "get_project_directory", "get_user_directory"]


def helpers_for(app_name: str) -> tuple[Callable[..., Any], Callable[..., None]]:
    """Return setting helpers bound to *app_name*.

    The returned ``get_setting`` and ``set_setting`` functions operate on a
    dedicated :class:`Sigil` instance for the provided application name.
    """

    sigil = Sigil(app_name)

    def get_setting(
        key: str,
        *,
        cast: Callable[[str], Any] | None = None,
        default: Any | None = None,
    ) -> Any:
        return sigil.get_pref(key, cast=cast, default=default)

    def set_setting(
        key: str,
        value: Any,
        *,
        scope: str | None = None,
    ) -> None:
        sigil.set_pref(key, value, scope=scope)

    return get_setting, set_setting


def get_project_directory(package_name: str) -> Path:
    """Return the root directory for an importable *package_name*."""

    return _paths.project_dir(package_name)


def get_user_directory(app_name: str) -> Path:
    """Return the per-user data directory for *app_name*."""

    return _paths.user_data_dir(app_name)



from __future__ import annotations

import importlib
import os
from importlib import resources as ilr
from pathlib import Path

from platformdirs import (
    user_cache_dir as _ucache,
    user_config_dir as _uc,
    user_data_dir as _ud,
)

from .root import ProjectRootNotFoundError, find_project_root

# ---------------------------------------------------------------------------
# User directories
# ---------------------------------------------------------------------------

def _app_name(default: str) -> str:
    return os.getenv("SIGIL_APP_NAME", default)

def user_config_dir(app_name: str = "pysigil") -> Path:
    app = _app_name(app_name)
    return Path(_uc(appname=app)).resolve()

def user_data_dir(app_name: str = "pysigil") -> Path:
    app = _app_name(app_name)
    return Path(_ud(appname=app)).resolve()

def user_cache_dir(app_name: str = "pysigil") -> Path:
    app = _app_name(app_name)
    return Path(_ucache(appname=app)).resolve()

# ---------------------------------------------------------------------------
# Project directories
# ---------------------------------------------------------------------------
def project_dir(package: str) -> Path:
    """Return the project root directory for an importable *package*."""

    module = importlib.import_module(package)
    file_attr = getattr(module, "__file__", None)
    if file_attr is not None:
        pkg_root = Path(file_attr).resolve().parent
    else:  # pragma: no cover - namespace package fallback
        pkg_root = Path(str(ilr.files(package))).resolve()
    try:
        root = find_project_root(start=pkg_root)
    except ProjectRootNotFoundError:
        root = pkg_root
    return root.resolve()


def project_root(start: str | Path | None = None, **kw) -> Path:
    env = os.getenv("SIGIL_ROOT")
    if env:
        return Path(env).expanduser().resolve()
    return find_project_root(start=start, **kw)

def project_config_dir(start: str | Path | None = None, **kw) -> Path:
    return project_root(start, **kw) / ".sigil"

def project_data_dir(start: str | Path | None = None, **kw) -> Path:
    return project_config_dir(start, **kw) / "data"

def project_cache_dir(start: str | Path | None = None, **kw) -> Path:
    return project_config_dir(start, **kw) / "cache"

# ---------------------------------------------------------------------------
# Package-installed defaults
# ---------------------------------------------------------------------------

def default_config_dir(package: str = "pysigil") -> Path:
    return Path(ilr.files(package)) / ".sigil"

def default_data_dir(package: str = "pysigil") -> Path:
    return default_config_dir(package) / "data"



from __future__ import annotations

import argparse
import configparser
import json
import os
import subprocess
import sys
from pathlib import Path

from .authoring import (
    DefaultsValidationError,
    DevLinkError,
    ensure_sigil_package_data,
    link as dev_link,
    list_links as dev_list,
    normalize_provider_id,
    unlink as dev_unlink,
    validate_defaults_file,
)
from .config import (
    ensure_gitignore as cfg_ensure_gitignore,
    host_file as cfg_host_file,
    init_config as cfg_init_config,
    load as cfg_load,
    open_scope as cfg_open_scope,
)
from .core import Sigil
from .discovery import pep503_name
from .errors import DevLinkNotFoundError
from .paths import (
    default_config_dir,
    default_data_dir,
    project_cache_dir,
    project_config_dir,
    project_data_dir,
    project_root as paths_project_root,
    user_cache_dir,
    user_config_dir,
    user_data_dir,
)
from .resolver import (
    default_provider_id,
    ensure_defaults_file,
    find_package_dir,
    read_dist_name_from_pyproject,
    validate_package_dir,
)
from .root import ProjectRootNotFoundError, find_project_root
from .ui.tk import launch as launch_gui

AUTHOR_FLAG_ENV = "SIGIL_AUTHOR"


def author_mode_enabled(args: argparse.Namespace | None = None) -> bool:
    """Return True if author mode should be enabled."""
    if args and getattr(args, "author", False):
        return True
    if os.environ.get(AUTHOR_FLAG_ENV):
        return True
    cfg = Path.home() / ".sigil" / "author.toml"
    return cfg.is_file()


def _launch(path: Path) -> None:  # pragma: no cover - best effort
    try:
        if sys.platform.startswith("win"):
            os.startfile(path)  # type: ignore[attr-defined]
        elif sys.platform == "darwin":
            subprocess.run(["open", str(path)], check=False)
        else:
            subprocess.run(["xdg-open", str(path)], check=False)
    except Exception:
        pass


# ---------------------------------------------------------------------------
# Config commands
# ---------------------------------------------------------------------------


def show_paths(args: argparse.Namespace) -> int:
    data = {
        "user_config": user_config_dir(),
        "user_data": user_data_dir(),
        "user_cache": user_cache_dir(),
        "project_root": paths_project_root(start=args.start),
        "project_config": project_config_dir(start=args.start),
        "project_data": project_data_dir(start=args.start),
        "project_cache": project_cache_dir(start=args.start),
        "default_config": default_config_dir(),
        "default_data": default_data_dir(),
    }
    if args.as_json:
        print(json.dumps({k: str(v) for k, v in data.items()}))
    else:
        for k, v in data.items():
            print(f"{k}: {v}")
    return 0


def config_init(args: argparse.Namespace) -> int:
    path = cfg_init_config(args.provider, args.scope, auto=args.auto)
    print(str(path))
    return 0


def config_open(args: argparse.Namespace) -> int:  # pragma: no cover - best effort
    path = cfg_open_scope(args.provider, args.scope, auto=args.auto)
    _launch(path)
    print(str(path))
    return 0


def config_host(args: argparse.Namespace) -> int:  # pragma: no cover - best effort
    try:
        path = cfg_host_file(args.provider, args.scope, auto=args.auto)
    except ValueError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    _launch(path)
    print(str(path))
    return 0


def config_show(args: argparse.Namespace) -> int:
    data = cfg_load(args.provider, auto=args.auto)
    if args.format == "json":
        print(json.dumps(data))
        return 0
    parser = configparser.ConfigParser()
    parser[normalize_provider_id(args.provider)] = {k: str(v) for k, v in data.items()}
    from io import StringIO

    buf = StringIO()
    parser.write(buf)
    print(buf.getvalue().strip())
    return 0


def config_gitignore(args: argparse.Namespace) -> int:
    if args.init:
        path = cfg_ensure_gitignore(auto=args.auto)
        print(str(path))
    return 0


def config_gui(_: argparse.Namespace) -> int:  # pragma: no cover - GUI interactions
    from .ui.tk.config_gui import launch as launch_config_gui

    launch_config_gui()
    return 0


# ---------------------------------------------------------------------------
# Basic commands
# ---------------------------------------------------------------------------


def get_cmd(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    val = sigil.get_pref(args.key)
    if val is None:
        return 1
    print(val)
    return 0


def set_cmd(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    sigil.set_pref(args.key, args.value, scope=args.scope)
    return 0


def secret_get(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    val = sigil.get_pref(args.key)
    if val is None:
        return 1
    print(val if args.reveal else "*" * 8)
    return 0


def secret_set(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    try:
        sigil.set_pref(args.key, args.value)
    except Exception:
        return 1
    return 0


def secret_unlock(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    sigil._secrets.unlock()
    return 0


def export_cmd(args: argparse.Namespace) -> int:
    sigil = Sigil(args.app)
    mapping = sigil.export_env(prefix=args.prefix)
    if args.as_json:
        print(json.dumps(mapping))
    else:
        for k, v in mapping.items():
            print(f"{k}={v}")
    return 0


def gui_cmd(args: argparse.Namespace) -> int:  # pragma: no cover - GUI interactions
    launch_gui(initial_provider=args.app, author_mode=author_mode_enabled(args))
    return 0


def launch_author_ui(_ctx: object) -> int:  # pragma: no cover - GUI interactions
    """Launch the author tools UI for the given context."""

    from .ui.core import AppCore
    from .ui.tk.author_tools import AuthorTools

    import tkinter as tk

    provider_id = getattr(_ctx, "provider_id", None)
    dev_root = getattr(_ctx, "dev_root", None)
    mode = getattr(_ctx, "mode", "edit")

    if dev_root is not None:
        os.environ.setdefault("SIGIL_ROOT", str(dev_root))

    core = AppCore(author_mode=True)
    if provider_id is not None:
        if mode == "bootstrap":
            core.orchestrator.register_provider(provider_id)
        try:
            core.select_provider(provider_id).result()
        except Exception:
            return 2

    root = tk.Tk()
    root.withdraw()
    win = AuthorTools(root, core)
    win.protocol("WM_DELETE_WINDOW", root.destroy)
    root.mainloop()
    return 0


def author_gui_cmd(_: argparse.Namespace) -> int:  # pragma: no cover - GUI interactions
    import sys

    from .ui.core import AppCore

    try:
        proj_root = find_project_root()
    except ProjectRootNotFoundError:
        proj_root = Path.cwd()
    dist_name = read_dist_name_from_pyproject(proj_root)
    pkg = find_package_dir(proj_root, dist_name)
    if not pkg:
        print("Could not auto-detect package directory", file=sys.stderr)
        return 2
    provider_id = default_provider_id(pkg, dist_name)

    core = AppCore(author_mode=True)
    pid = normalize_provider_id(provider_id)
    try:
        ctx = core.orchestrator.load_author_context(pid)
    except DevLinkNotFoundError:
        err = (
            f"No development link for '{pid}'.\n"
            f"Create one:\n  sigil link --dev /path/to/pkg --provider {pid}"
        )
        print(err, file=sys.stderr)
        return 2
    except Exception as exc:
        print(f"Failed to prepare authoring for '{pid}': {exc}", file=sys.stderr)
        return 2

    return launch_author_ui(ctx)


def setup_cmd(_: argparse.Namespace) -> int:  # pragma: no cover - GUI interactions
    from .ui.tk.author import main as author_main

    author_main()
    return 0


# ---------------------------------------------------------------------------
# Author commands
# ---------------------------------------------------------------------------


def author_register(args: argparse.Namespace) -> int:
    if args.auto:
        try:
            root = find_project_root()
        except ProjectRootNotFoundError:
            root = Path.cwd()
        dist_name = read_dist_name_from_pyproject(root)
        pkg = find_package_dir(root, dist_name)
        if not pkg:
            print("Could not auto-detect package directory", file=sys.stderr)
            return 2
        try:
            pkg = validate_package_dir(pkg)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        import_name = pkg.name
        provider_id = (
            normalize_provider_id(args.provider)
            if args.provider
            else default_provider_id(pkg, dist_name)
        )
        settings_path = ensure_defaults_file(pkg, provider_id)
        ensure_sigil_package_data(root, import_name)
        try:
            dev_link(provider_id, settings_path, validate=not args.no_validate)
        except DevLinkError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        print(f"Linked {provider_id} -> {settings_path}")
        return 0

    if bool(args.package_dir) == bool(args.defaults):
        print("Use exactly one of --package-dir or --defaults", file=sys.stderr)
        return 2

    if args.package_dir:
        try:
            pkg = validate_package_dir(args.package_dir)
        except ValueError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        try:
            root = find_project_root(pkg)
        except ProjectRootNotFoundError:
            root = pkg
        dist_name = read_dist_name_from_pyproject(root)
        provider_id = (
            normalize_provider_id(args.provider)
            if args.provider
            else default_provider_id(pkg, dist_name)
        )
        settings_path = ensure_defaults_file(pkg, provider_id)
        ensure_sigil_package_data(root, pkg.name)
        try:
            dev_link(provider_id, settings_path, validate=not args.no_validate)
        except DevLinkError as exc:
            print(str(exc), file=sys.stderr)
            return 2
        print(f"Linked {provider_id} -> {settings_path}")
        return 0

    assert args.defaults is not None
    if not args.provider:
        print("--provider is required when using --defaults", file=sys.stderr)
        return 2
    provider_id = normalize_provider_id(args.provider)
    try:
        dev_link(provider_id, args.defaults, validate=not args.no_validate)
    except DevLinkError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    print(f"Linked {provider_id} -> {args.defaults}")
    return 0


def author_link_defaults(args: argparse.Namespace) -> int:
    provider = pep503_name(args.provider_id)
    try:
        dev_link(provider, args.path)
    except (DefaultsValidationError, DevLinkError):
        return 1
    return 0


def author_unlink_defaults(args: argparse.Namespace) -> int:
    provider = pep503_name(args.provider_id)
    ok = dev_unlink(provider)
    if not ok:
        return 1
    return 0


def author_validate(args: argparse.Namespace) -> int:
    provider = pep503_name(args.provider_id)
    try:
        validate_defaults_file(args.path, provider)
    except DefaultsValidationError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    return 0


def author_list_links(args: argparse.Namespace) -> int:
    entries = dev_list(must_exist_on_disk=args.existing_only)
    if not entries:
        print("No dev links found")
        return 0
    for pid, path in sorted(entries.items()):
        status = "(ok)" if path.exists() else "(missing)"
        print(f"{pid}: {path} {status}")
    return 0


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="sigil", description="Sigil command line interface.")
    parser.add_argument("--author", action="store_true", help="Enable author mode")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # paths command
    p_paths = subparsers.add_parser("paths", help="Show Sigil paths.")
    p_paths.add_argument("--start", type=Path, default=None)
    p_paths.add_argument("--json", dest="as_json", action="store_true")
    p_paths.set_defaults(func=show_paths)

    # config group
    p_config = subparsers.add_parser("config", help="Manage Sigil configuration files.")
    sp_config = p_config.add_subparsers(dest="config_cmd", required=True)

    p_cfg_init = sp_config.add_parser("init", help="Initialize configuration")
    p_cfg_init.add_argument("--provider", required=True)
    p_cfg_init.add_argument("--scope", choices=["user", "project"], default="user")
    p_cfg_init.add_argument("--auto", action="store_true", help="Auto-detect project root")
    p_cfg_init.set_defaults(func=config_init)

    p_cfg_open = sp_config.add_parser("open", help="Open configuration scope")
    p_cfg_open.add_argument("--provider", required=True)
    p_cfg_open.add_argument("--scope", choices=["user", "project"], default="user")
    p_cfg_open.add_argument("--auto", action="store_true", help="Auto-detect project root")
    p_cfg_open.set_defaults(func=config_open)

    p_cfg_host = sp_config.add_parser("host", help="Open host-specific config")
    p_cfg_host.add_argument("--provider", required=True)
    p_cfg_host.add_argument("--scope", choices=["user", "project"], default="user")
    p_cfg_host.add_argument("--auto", action="store_true", help="Auto-detect project root")
    p_cfg_host.set_defaults(func=config_host)

    p_cfg_show = sp_config.add_parser("show", help="Show configuration")
    p_cfg_show.add_argument("--provider", required=True)
    p_cfg_show.add_argument("--as", dest="format", choices=["ini", "json"], default="ini")
    p_cfg_show.add_argument("--auto", action="store_true", help="Auto-detect project root")
    p_cfg_show.set_defaults(func=config_show)

    p_cfg_git = sp_config.add_parser("gitignore", help="Manage .gitignore")
    p_cfg_git.add_argument("--init", action="store_true", help="Add ignore rule")
    p_cfg_git.add_argument("--auto", action="store_true", help="Auto-detect project root")
    p_cfg_git.set_defaults(func=config_gitignore)

    p_cfg_gui = sp_config.add_parser("gui", help="Launch the config GUI")
    p_cfg_gui.set_defaults(func=config_gui)

    # get command
    p_get = subparsers.add_parser("get", help="Print the value for KEY.")
    p_get.add_argument("key")
    p_get.add_argument("--app", required=True)
    p_get.set_defaults(func=get_cmd)

    # set command
    p_set = subparsers.add_parser("set", help="Set KEY to VALUE.")
    p_set.add_argument("key")
    p_set.add_argument("value")
    p_set.add_argument("--app", required=True)
    p_set.add_argument("--scope", choices=["user", "project"], default="user")
    p_set.set_defaults(func=set_cmd)

    # secret group
    p_secret = subparsers.add_parser("secret", help="Manage secret preferences.")
    sp_secret = p_secret.add_subparsers(dest="secret_cmd", required=True)

    p_sec_get = sp_secret.add_parser("get", help="Get secret value")
    p_sec_get.add_argument("key")
    p_sec_get.add_argument("--app", required=True)
    p_sec_get.add_argument("--reveal", action="store_true", help="Print the secret value")
    p_sec_get.set_defaults(func=secret_get)

    p_sec_set = sp_secret.add_parser("set", help="Set secret value")
    p_sec_set.add_argument("key")
    p_sec_set.add_argument("value")
    p_sec_set.add_argument("--app", required=True)
    p_sec_set.set_defaults(func=secret_set)

    p_sec_unlock = sp_secret.add_parser("unlock", help="Unlock secret store")
    p_sec_unlock.add_argument("--app", required=True)
    p_sec_unlock.set_defaults(func=secret_unlock)

    # export command
    p_export = subparsers.add_parser("export", help="Export preferences as environment variables.")
    p_export.add_argument("--app", required=True)
    p_export.add_argument("--prefix", default="SIGIL_")
    p_export.add_argument("--json", dest="as_json", action="store_true")
    p_export.set_defaults(func=export_cmd)

    # gui command
    p_gui = subparsers.add_parser("gui", help="Launch the preferences GUI.")
    p_gui.add_argument("--app")
    p_gui.add_argument("--include-sigil", action="store_true")
    p_gui.add_argument("--no-remember", action="store_true")
    p_gui.set_defaults(func=gui_cmd)

    # setup command (alias: register)
    p_setup = subparsers.add_parser(
        "setup", help="Launch the defaults registration GUI.", aliases=["register"]
    )
    p_setup.set_defaults(func=setup_cmd)

    # author command / group
    p_author = subparsers.add_parser(
        "author", help="Launch the author tools GUI or manage development links."
    )
    p_author.set_defaults(func=author_gui_cmd)
    sp_author = p_author.add_subparsers(dest="author_cmd")

    p_reg = sp_author.add_parser("register", help="Register package defaults for development.")
    p_reg.add_argument("--package-dir", type=Path, help="Package directory")
    p_reg.add_argument("--defaults", type=Path, help="Path to settings.ini")
    p_reg.add_argument("--provider", help="Override provider id")
    p_reg.add_argument("--auto", action="store_true", help="Auto-detect from current directory")
    p_reg.add_argument("--no-validate", action="store_true", help="Skip defaults validation")
    p_reg.set_defaults(func=author_register)

    p_link = sp_author.add_parser("link-defaults", help="Link defaults file for provider")
    p_link.add_argument("provider_id")
    p_link.add_argument("path", type=Path)
    p_link.set_defaults(func=author_link_defaults)

    p_unlink = sp_author.add_parser("unlink-defaults", help="Unlink defaults file for provider")
    p_unlink.add_argument("provider_id")
    p_unlink.set_defaults(func=author_unlink_defaults)

    p_val = sp_author.add_parser("validate", help="Validate defaults file")
    p_val.add_argument("provider_id")
    p_val.add_argument("path", type=Path)
    p_val.set_defaults(func=author_validate)

    p_list = sp_author.add_parser("list", help="List dev links")
    p_list.add_argument("--existing-only", action="store_true")
    p_list.set_defaults(func=author_list_links)

    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    func = getattr(args, "func", None)
    if func is None:
        parser.print_help()
        return 1
    try:
        return int(func(args))
    except SystemExit as exc:  # pragma: no cover - argparse may raise
        return int(exc.code)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())




"""Build per-row scope menu data structures.

This module contains a small helper that translates the specification for
the per-row scope dropdown into a framework agnostic representation.  The
resulting list can be consumed by concrete UI toolkits to render the menu
and wire callbacks.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List

import pysigil

from .. import api


@dataclass
class Action:
    label: str
    enabled: bool = True
    handler: Callable[[], Any] | None = None


@dataclass
class ScopeRow:
    scope: str
    effective: bool = False
    present: bool = False
    actions: List[Action] = field(default_factory=list)

    def add(self, action: Action) -> None:
        self.actions.append(action)


@dataclass
class Separator:
    pass


def build_menu(
    handle: api.ProviderHandle,
    key: str,
    active_scope: str,
    policy,
) -> List[object]:
    """Return a list describing the per-row scope menu.

    The returned list contains :class:`Action`, :class:`ScopeRow` and
    :class:`Separator` instances as described in the specification.  Concrete
    view layers can interpret this structure to render the dropdown.
    """

    eff = handle.effective().get(key)
    layers = handle.layers().get(key, {})
    scopes = ["user", "user-local", "project", "project-local"]
    if not pysigil.show_machine_scope:
        machine = set(policy.machine_scopes())
        scopes = [s for s in scopes if s not in machine]

    paths = {scope: handle.target_path(scope) for scope in scopes}

    def has(scope: str) -> bool:
        return layers.get(scope) is not None

    def writable(scope: str) -> bool:
        allows = getattr(policy, "allows", lambda s: True)
        return bool(allows(scope))

    menu: List[object] = []

    # Quick actions
    menu.append(Action(f"Edit at {active_scope}", enabled=writable(active_scope)))
    same_as_eff = has(active_scope) and eff is not None and layers[active_scope].value == eff.value
    copy_enabled = writable(active_scope) and eff is not None and eff.value is not None and not same_as_eff
    menu.append(
        Action(
            f"Copy effective here ({active_scope})",
            enabled=copy_enabled,
        )
    )
    menu.append(Separator())

    # Scopes list
    for scope in scopes:
        row = ScopeRow(
            scope,
            effective=(eff is not None and eff.source == scope),
            present=has(scope),
        )
        if has(scope):
            row.add(Action("Edit…", enabled=writable(scope)))
            row.add(Action("Remove", enabled=writable(scope)))
        else:
            row.add(Action("Add…", enabled=writable(scope)))
        menu.append(row)

    menu.append(Separator())
    menu.append(Action("Show layers…"))

    # Files
    for scope in [s for s in ["user", "project", "user-local", "project-local"] if s in paths]:
        menu.append(
            Action(f"Open file for {scope}", enabled=paths.get(scope) is not None)
        )

    return menu


__all__ = ["Action", "ScopeRow", "Separator", "build_menu"]




from __future__ import annotations

import os
import re
from collections.abc import MutableMapping

KeyPath = tuple[str, ...]

_SPLIT_RX = re.compile(r"[._]")


def parse_key(raw: str | KeyPath) -> KeyPath:
    if isinstance(raw, tuple):
        return raw
    parts = _SPLIT_RX.split(raw) if _SPLIT_RX.search(raw) else [raw]
    if any(p == "" for p in parts):
        raise ValueError(f"Malformed key '{raw}'")
    return tuple(parts)


def read_env(app_name: str) -> MutableMapping[KeyPath, str]:
    sanitized = app_name.upper().replace("-", "_")
    prefix = f"SIGIL_{sanitized}_"
    result: MutableMapping[KeyPath, str] = {}
    for key, value in os.environ.items():
        if key.startswith(prefix):
            raw = key[len(prefix):].lower()
            result[parse_key(raw)] = value
    return result




try:  # pragma: no cover - tkinter availability depends on the env
    import tkinter as tk
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore

import pytest

from pysigil.ui.tk import App


def _make_app(author_mode: bool) -> App:
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    return App(root, author_mode=author_mode)


def test_author_tools_requires_author_mode():
    app = _make_app(author_mode=False)
    with pytest.raises(RuntimeError):
        app._open_author_tools()
    app.root.destroy()


def test_author_tools_window_opens_in_author_mode():
    app = _make_app(author_mode=True)
    app._open_author_tools()
    assert app._author_tools is not None
    assert bool(app._author_tools.winfo_exists())
    app._author_tools.destroy()
    app.root.destroy()



from __future__ import annotations

import importlib
import os
import sys
from pathlib import Path


from pysigil import (
    get_project_directory,
    get_user_directory,
    helpers_for,
)


def test_helpers_for_isolated_apps(monkeypatch, tmp_path: Path) -> None:
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
    monkeypatch.delenv("SIGIL_DEMO_SECTION_VALUE", raising=False)

    get_a, set_a = helpers_for("demo")
    set_a("section.value", "1")
    assert get_a("section.value", cast=int) == 1

    get_b, set_b = helpers_for("other")
    assert get_b("section.value") is None
    set_b("section.value", "2")
    assert get_b("section.value", cast=int) == 2

    # original demo settings remain unchanged
    assert get_a("section.value", cast=int) == 1


def test_helpers_environment_scope(monkeypatch, tmp_path: Path) -> None:
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))

    get_setting, set_setting = helpers_for("demo-env")
    env_key = "SIGIL_DEMO_ENV_SECTION_VALUE"
    monkeypatch.delenv(env_key, raising=False)

    try:
        set_setting("section.value", "7", scope="environment")
        assert os.environ[env_key] == "7"
        assert get_setting("section.value") == 7
    finally:
        set_setting("section.value", None, scope="env")
    assert env_key not in os.environ
    assert get_setting("section.value") is None


def test_get_user_directory_is_app_specific(monkeypatch, tmp_path: Path) -> None:
    monkeypatch.delenv("SIGIL_APP_NAME", raising=False)

    calls: list[str] = []

    def fake_user_data_dir(*, appname: str) -> Path:
        calls.append(appname)
        return tmp_path / f"data-{appname}"

    monkeypatch.setattr("pysigil.paths._ud", fake_user_data_dir)

    first = get_user_directory("alpha-app")
    second = get_user_directory("beta-app")

    assert first == (tmp_path / "data-alpha-app").resolve()
    assert second == (tmp_path / "data-beta-app").resolve()
    assert first.is_absolute()
    assert second.is_absolute()
    assert calls == ["alpha-app", "beta-app"]


def test_get_project_directory_returns_package_root(
    monkeypatch, tmp_path: Path
) -> None:
    names = ("toolkit_pkg_alpha", "toolkit_pkg_beta")
    roots = {}
    for name in names:
        root = tmp_path / name
        pkg_dir = root / name
        pkg_dir.mkdir(parents=True)
        (pkg_dir / "__init__.py").write_text("VALUE = 1\n", encoding="utf-8")
        (root / "pyproject.toml").write_text("[project]\nname='demo'\n", encoding="utf-8")
        monkeypatch.syspath_prepend(str(root))
        roots[name] = root.resolve()

    importlib.invalidate_caches()

    try:
        resolved = {name: get_project_directory(name) for name in names}
    finally:
        for name in names:
            sys.modules.pop(name, None)

    assert resolved["toolkit_pkg_alpha"].is_absolute()
    assert resolved["toolkit_pkg_beta"].is_absolute()
    assert resolved["toolkit_pkg_alpha"] == roots["toolkit_pkg_alpha"]
    assert resolved["toolkit_pkg_beta"] == roots["toolkit_pkg_beta"]
    assert resolved["toolkit_pkg_alpha"] != resolved["toolkit_pkg_beta"]




from .core import Sigil, SigilError
from .merge_policy import parse_key
from .policy import policy
from .toolkit import get_project_directory, get_user_directory, helpers_for
from .root import find_project_root

# Toggle visibility of machine-specific scopes in the UI.  When ``False``
# machine scopes such as ``user-local`` and ``project-local`` are hidden.
show_machine_scope = False


__all__ = [
    "Sigil",
    "SigilError",
    "parse_key",
    "policy",
    "helpers_for",
    "get_project_directory",
    "get_user_directory",
    "show_machine_scope",
]



from __future__ import annotations

import pytest

from pysigil.policy import (
    PRECEDENCE_PROJECT_WINS,
    PRECEDENCE_USER_WINS,
    SigilWriteError,
    policy as default_policy,
)


def test_precedence_modes() -> None:
    p = default_policy.clone()
    assert p.precedence(read=True) == PRECEDENCE_PROJECT_WINS

    store = {("pysigil", "policy"): "user_over_project"}
    p.set_store("user", store)
    assert p.precedence(read=True) == PRECEDENCE_USER_WINS
    assert p.precedence(read=False) == PRECEDENCE_PROJECT_WINS


def test_write_rules(tmp_path, monkeypatch) -> None:
    from pysigil import config as cfg

    monkeypatch.setattr(cfg, "user_config_dir", lambda app: str(tmp_path))
    monkeypatch.setattr(cfg, "_project_dir", lambda auto: tmp_path)

    p = default_policy.clone()

    path = p.path("user", "mypkg")
    assert path == tmp_path / "mypkg" / "settings.ini"

    with pytest.raises(SigilWriteError):
        p.path("env", "mypkg")



from __future__ import annotations

import configparser
import logging
import socket
from pathlib import Path
from typing import Any

from .paths import user_config_dir

from .authoring import normalize_provider_id
from .root import ProjectRootNotFoundError, find_project_root
from .policy import policy
from .io_config import IniIOError


logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Host and provider helpers
# ---------------------------------------------------------------------------

def host_id() -> str:
    """Return the normalised hostname."""
    raw = socket.gethostname()
    return normalize_provider_id(raw).strip("-")


# ---------------------------------------------------------------------------
# File discovery
# ---------------------------------------------------------------------------

def user_files(provider_id: str, host: str) -> list[Path]:
    base = Path(user_config_dir("sigil")) / provider_id
    files = [base / "settings.ini", base / f"settings-local-{host}.ini"]
    return [f for f in files if f.exists()]


def _project_dir(auto: bool) -> Path | None:
    if auto:
        try:
            return find_project_root()
        except ProjectRootNotFoundError:
            return None
    return Path.cwd()


def project_files(provider_id: str, host: str, *, auto: bool = True) -> list[Path]:
    root = _project_dir(auto)
    if root is None:
        return []
    base = root / ".sigil"
    files = [base / "settings.ini", base / f"settings-local-{host}.ini"]
    return [f for f in files if f.exists()]


# ---------------------------------------------------------------------------
# Reading helpers
# ---------------------------------------------------------------------------

def merge_ini_section(acc: dict[str, Any], ini_path: Path, *, section: str) -> dict[str, Any]:
    parser = configparser.ConfigParser()
    try:
        parser.read(ini_path)
    except Exception as exc:  # pragma: no cover - defensive
        raise IniIOError(str(exc)) from exc
    if parser.has_section(section):
        for k, v in parser.items(section):
            acc[k] = v
    return acc


def load(provider_id: str, *, auto: bool = True) -> dict[str, Any]:
    pid = normalize_provider_id(provider_id)
    acc: dict[str, Any] = {}
    for scope in reversed(policy.precedence(read=True)):
        if scope not in {"user", "user-local", "project", "project-local"}:
            continue
        try:
            path = policy.path(scope, pid, auto=auto)
        except ProjectRootNotFoundError:
            continue
        if not path.exists():
            continue
        try:
            acc = merge_ini_section(acc, path, section=pid)
        except IniIOError as exc:
            logger.warning("Failed to read config %s: %s", path, exc)
    return acc


# ---------------------------------------------------------------------------
# Writing helpers used by CLI and GUI
# ---------------------------------------------------------------------------

def _seed_section(path: Path, section: str, comment: str) -> None:
    if path.exists():
        parser = configparser.ConfigParser()
        try:
            parser.read(path)
        except Exception as exc:  # pragma: no cover - defensive
            logger.warning("Failed to read config %s: %s", path, exc)
            parser = None
        if parser and parser.has_section(section):
            return
        with path.open("a") as fh:
            fh.write(f"\n[{section}]\n{comment}")
        return
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(f"[{section}]\n{comment}")


def init_config(provider_id: str, scope: str, *, auto: bool = False) -> Path:
    pid = normalize_provider_id(provider_id)
    path = policy.path(scope, pid, auto=auto)
    comment = (
        "# per-machine settings\n"
        if scope in policy.machine_scopes() or pid == "user-custom"
        else "# add keys here\n"
    )
    _seed_section(path, pid, comment)
    return path


def target_path(provider_id: str, scope: str, *, auto: bool = False) -> Path:
    """Return the configuration file path for *provider_id* and *scope*."""

    return policy.path(scope, provider_id, auto=auto)


def open_scope(provider_id: str, scope: str, *, auto: bool = False) -> Path:
    return policy.path(scope, provider_id, auto=auto).parent


def host_file(provider_id: str, scope: str, *, auto: bool = False) -> Path:
    pid = normalize_provider_id(provider_id)
    if pid != "user-custom":
        raise ValueError("host command is only valid for provider 'user-custom'")
    return init_config(pid, scope, auto=auto)


def ensure_gitignore(*, auto: bool = False) -> Path:
    root = _project_dir(auto)
    if root is None:
        raise ProjectRootNotFoundError("No project root found")
    gi = root / ".gitignore"
    lines: list[str] = []
    if gi.exists():
        lines = gi.read_text().splitlines()
    if "project-local" in policy.machine_scopes():
        path = policy.path("project-local", "user-custom", auto=auto)
        rule = str(path.relative_to(root).parent / "settings-local*")
        if rule not in lines:
            lines.append(rule)
            gi.write_text("\n".join(lines) + "\n")
    return gi


def available_providers(*, auto: bool = True) -> list[str]:
    """Return provider IDs present in user or project config directories."""

    providers: set[str] = set()
    user_base = Path(user_config_dir("sigil"))
    if user_base.exists():
        for p in user_base.iterdir():
            if p.is_dir():
                providers.add(normalize_provider_id(p.name))
    root = _project_dir(auto)
    if root is not None:
        proj_base = root / ".sigil"
        if proj_base.exists():
            for ini in proj_base.glob("settings*.ini"):
                parser = configparser.ConfigParser()
                try:
                    parser.read(ini)
                except Exception as exc:  # pragma: no cover - defensive
                    logger.warning("Failed to read config %s: %s", ini, exc)
                    continue
                for section in parser.sections():
                    providers.add(normalize_provider_id(section))
    return sorted(providers)




# pysigil/root.py
from __future__ import annotations

import re
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path

__all__ = [
    "ProjectRootNotFoundError",
    "ProjectRootNotFoundWithSuggestionsError",
    "Candidate",
    "find_project_root",
    "suggest_candidates",
]

# =========================================================
# Exceptions & suggestion model
# =========================================================

class ProjectRootNotFoundError(RuntimeError):
    """Raised when no project root can be located."""

@dataclass(frozen=True)
class Candidate:
    """A ranked fallback directory that *looks* like a project root."""
    path: Path
    score: int
    hits: tuple[str, ...]  # marker names contributing to the score

class ProjectRootNotFoundWithSuggestionsError(ProjectRootNotFoundError):
    """
    Raised when no strong root is found; includes ranked candidates.
    Catch this to present choices and (optionally) place a sentinel.
    """
    def __init__(self, start: Path, candidates: list[Candidate]):
        super().__init__(f"No project root found from {start}")
        self.start = start
        self.candidates = candidates

# Backwards compatibility
ProjectRootNotFoundWithSuggestions = ProjectRootNotFoundWithSuggestionsError

# =========================================================
# Hard-coded strong tiers (auto-detect)
# =========================================================

# Tier 0 – explicit sentinel (owner intent beats all)
_SENTINEL_FILE = ".sigil-root"

# Tier 1 – Python packaging + VCS
_TIER1_FILES: tuple[str, ...] = ("pyproject.toml", "setup.cfg", "setup.py")
_TIER1_DIRS:  tuple[str, ...] = (".git", ".hg", ".svn")

# Tier 2 – Python tools
_TIER2_FILES: tuple[str, ...] = ("tox.ini", "poetry.lock", "pdm.lock")
_TIER2_DIRS:  tuple[str, ...] = (".dvc",)

# Tier 3 – other ecosystems / monorepo roots
_TIER3_FILES: tuple[str, ...] = (
    # Bazel/CMake
    "WORKSPACE", "WORKSPACE.bazel", "MODULE.bazel", "CMakeLists.txt",
    # JS/TS monorepos & tools
    "package.json", "pnpm-workspace.yaml", "lerna.json", "turbo.json", "nx.json",
    # Go / Rust / Java/Kotlin
    "go.mod", "Cargo.toml", "pom.xml", "settings.gradle", "settings.gradle.kts",
)
_TIER3_DIRS: tuple[str, ...] = (".github",)  # weak but repo-level and useful

# IDE dirs (treated as strong by default, but with guards)
_IDE_DIRS: tuple[str, ...] = (".idea", ".vscode")

# =========================================================
# Weak “massive net” (suggestions only; never auto-pick)
# =========================================================

_WEAK_FILES: tuple[str, ...] = (
    "README", "README.md", "README.rst",
    "LICENSE", "LICENSE.md",
    ".gitignore", ".gitattributes", ".editorconfig",
    ".pre-commit-config.yaml", ".ruff.toml", ".flake8",
    "requirements.txt", "requirements-dev.txt", "environment.yml",
    "Pipfile", "Pipfile.lock", ".python-version",
    "Makefile", "Dockerfile", "docker-compose.yml", "compose.yaml",
    ".gitlab-ci.yml", "azure-pipelines.yml", "Jenkinsfile",
    ".nvmrc", ".node-version", ".tool-versions", "tsconfig.json",
)
_WEAK_DIRS: tuple[str, ...] = (".github", ".github/workflows", ".idea", ".vscode")

_WEIGHTS: dict[str, int] = {
    # docs / legal
    "README": 10, "README.md": 10, "README.rst": 10,
    "LICENSE": 10, "LICENSE.md": 10,
    # git/editor hygiene
    ".gitignore": 10, ".gitattributes": 10, ".editorconfig": 8,
    # qa/tooling
    ".pre-commit-config.yaml": 8, ".ruff.toml": 8, ".flake8": 6,
    # python env
    "requirements.txt": 12, "requirements-dev.txt": 12, "environment.yml": 12,
    "Pipfile": 12, "Pipfile.lock": 12, ".python-version": 6,
    # build/devops
    "Makefile": 10, "Dockerfile": 10, "docker-compose.yml": 10, "compose.yaml": 10,
    # CI
    ".github": 10, ".github/workflows": 10, ".gitlab-ci.yml": 10,
    "azure-pipelines.yml": 10, "Jenkinsfile": 10,
    # language hints
    ".nvmrc": 6, ".node-version": 6, ".tool-versions": 6, "tsconfig.json": 6,
    # IDE dirs (weak as hints only)
    ".idea": 6, ".vscode": 6,
}

# =========================================================
# Internals
# =========================================================

_GITDIR_RE = re.compile(r"^\s*gitdir:\s*(.+)\s*$", re.IGNORECASE)

def _has_git_marker(d: Path) -> bool:
    g = d / ".git"
    if g.is_dir():
        return True
    if g.is_file():
        try:
            return bool(_GITDIR_RE.search(g.read_text(errors="ignore")))
        except Exception:
            return False
    return False

def _walk_up(start: Path) -> Iterable[Path]:
    cur = start
    while True:
        yield cur
        if cur.parent == cur:
            break
        cur = cur.parent

def _match_here(cur: Path, *, files: Sequence[str], dirs: Sequence[str]) -> str | None:
    for f in files:
        if (cur / f).is_file():
            return f
    for d in dirs:
        if d == ".git":
            if _has_git_marker(cur):
                return ".git"
        elif (cur / d).is_dir():
            return d
    return None

# ---------- IDE guard (prevents stray matches) ----------

def _is_real_vscode(cur: Path) -> bool:
    vs = cur / ".vscode"
    return any((vs / n).is_file() for n in ("settings.json", "launch.json", "tasks.json"))

def _is_real_jetbrains(cur: Path) -> bool:
    ij = cur / ".idea"
    return any((ij / n).is_file() for n in ("workspace.xml", "modules.xml", "misc.xml", "project.iml"))

_CO_SIGNALS: tuple[str, ...] = (
    "README", "README.md", "Makefile", "Dockerfile",
    "requirements.txt", "requirements-dev.txt", "environment.yml",
    ".gitignore", ".editorconfig", "package.json",
)

def _has_co_signal(cur: Path) -> bool:
    return any((cur / f).exists() for f in _CO_SIGNALS)

def _ide_guard_ok(cur: Path) -> bool:
    return _is_real_vscode(cur) or _is_real_jetbrains(cur) or _has_co_signal(cur)

# ---------- Suggestions scoring ----------

def _score_dir(cur: Path) -> Candidate | None:
    hits: list[str] = []
    score = 0
    for f in _WEAK_FILES:
        if (cur / f).is_file():
            hits.append(f)
            score += _WEIGHTS.get(f, 5)
    for d in _WEAK_DIRS:
        if (cur / d).is_dir():
            hits.append(d)
            score += _WEIGHTS.get(d, 5)
    if score == 0:
        return None
    return Candidate(path=cur, score=score, hits=tuple(hits))

# =========================================================
# Public API
# =========================================================

def suggest_candidates(
    start: str | Path | None = None,
    *,
    max_up: int = 16,
    top_k: int = 5,
) -> list[Candidate]:
    """
    Return top-k directories above `start` that *look* like project roots,
    ranked by weak-marker score (higher is better), then by proximity (nearer first).
    """
    start_path = (Path.cwd() if start is None else Path(start)).resolve(strict=False)

    # distance map for proximity tie-break
    dist_index: dict[Path, int] = {}
    for i, cur in enumerate(_walk_up(start_path)):
        dist_index[cur] = i
        if i >= max_up:
            break

    cands: list[Candidate] = []
    for cur, _i in dist_index.items():
        cand = _score_dir(cur)
        if cand:
            cands.append(cand)

    cands.sort(key=lambda c: (-c.score, dist_index.get(c.path, 1_000_000)))
    return cands[:top_k]

def find_project_root(
    start: str | Path | None = None,
    *,
    strict: bool = True,
) -> Path:
    """
    Find the nearest project root above `start` using strong markers.

    Priority (two-pass, nearest within each pass wins):
      Pass A:  .sigil-root  →  Tier1 (py packaging + VCS)  →  Tier2 (py tools)
      Pass B:  Tier3 (other ecosystems)  →  IDE dirs (guarded)

    If nothing matches and `strict=True`, raises
    ProjectRootNotFoundWithSuggestionsError containing ranked fallback
    candidates (weak 'massive net').
    If `strict=False`, returns the resolved starting directory.
    """
    start_path = (Path.cwd() if start is None else Path(start)).resolve(strict=False)

    # Pass A: strongest tiers
    for cur in _walk_up(start_path):
        if (cur / _SENTINEL_FILE).is_file():
            return cur
        if _match_here(cur, files=_TIER1_FILES, dirs=_TIER1_DIRS):
            return cur
        if _match_here(cur, files=_TIER2_FILES, dirs=_TIER2_DIRS):
            return cur

    # Pass B: other ecosystems (then guarded IDE markers)
    for cur in _walk_up(start_path):
        if _match_here(cur, files=_TIER3_FILES, dirs=_TIER3_DIRS):
            return cur
        if any((cur / d).is_dir() for d in _IDE_DIRS) and _ide_guard_ok(cur):
            return cur

    # Nothing strong found → suggest candidates or fallback
    if strict:
        raise ProjectRootNotFoundWithSuggestionsError(
            start_path, suggest_candidates(start_path)
        )
    return start_path



class SigilError(Exception):
    """Base class for Sigil errors."""


class UnknownScopeError(SigilError):
    pass


class SigilLoadError(SigilError):
    """Raised when a backend fails to parse its file."""


class SigilMetaError(SigilError):
    """Raised when preference metadata is malformed."""


class SigilWriteError(SigilError):
    """Raised when attempting to write and the target is read-only."""


class ReadOnlyScopeError(SigilError):
    """Raised when attempting to modify the read-only core scope."""


class SigilSecretsError(SigilError):
    """Raised for errors in the secrets subsystem."""


class UnknownProviderError(SigilError):
    """Raised when a provider is not registered."""


class DevLinkNotFoundError(SigilError):
    """Raised when a development link is missing."""


class DuplicateProviderError(SigilError):
    """Raised when attempting to create a provider that already exists."""


class ConflictError(SigilError):
    """Raised when concurrent modifications conflict."""


class UnknownFieldError(SigilError):
    """Raised when a field key is unknown."""


class DuplicateFieldError(SigilError):
    """Raised when attempting to add a field that already exists."""


class ValidationError(SigilError):
    """Raised when value validation fails."""


class PolicyError(SigilError):
    """Raised when policy prevents an operation."""


class IOFailureError(SigilError):
    """Raised for unexpected IO failures."""



import os
from types import SimpleNamespace
from pathlib import Path

import pytest

from pysigil.cli import build_parser, author_mode_enabled
from pysigil.ui.provider_adapter import ProviderAdapter
from pysigil.ui.author_adapter import AuthorAdapter
from pysigil.ui.value_parser import parse_field_value
from pysigil.errors import ValidationError
from pysigil import api, authoring
from pysigil.settings_metadata import IniFileBackend, IniSpecBackend
from pysigil.orchestrator import Orchestrator
from tests.utils import DummyPolicy

def test_author_mode_detection(tmp_path, monkeypatch):
    parser = build_parser()
    args = parser.parse_args(["--author", "paths"])
    assert author_mode_enabled(args)
    monkeypatch.delenv("SIGIL_AUTHOR", raising=False)
    args = parser.parse_args(["paths"])
    monkeypatch.setenv("SIGIL_AUTHOR", "1")
    assert author_mode_enabled(args)
    monkeypatch.delenv("SIGIL_AUTHOR")
    author_file = tmp_path / ".sigil" / "author.toml"
    author_file.parent.mkdir(parents=True)
    author_file.write_text("1")
    monkeypatch.setattr(Path, "home", lambda: tmp_path)
    assert author_mode_enabled(args)


def _adapter_with_default(tmp_path, monkeypatch, *, author_mode: bool) -> ProviderAdapter:
    monkeypatch.setattr(authoring, "user_config_dir", lambda app="sigil": tmp_path / "cfg")
    spec = IniSpecBackend(user_dir=tmp_path / "meta")
    pol = DummyPolicy(tmp_path / "user", tmp_path / "proj")
    cfg = IniFileBackend(policy=pol)
    orch = Orchestrator(spec_backend=spec, config_backend=cfg)
    monkeypatch.setattr(api, "_ORCH", orch, raising=False)
    pid = "demo-auth"
    api.register_provider(pid, title="Demo")
    handle = api.handle(pid)
    handle.add_field("alpha", "integer")
    defaults_dir = tmp_path / "pkg" / ".sigil"
    defaults_dir.mkdir(parents=True)
    defaults_ini = defaults_dir / "settings.ini"
    defaults_ini.write_text(f"[{pid}]\n")
    authoring.link(pid, defaults_ini, validate=False)
    adapter = ProviderAdapter(author_mode=author_mode)
    adapter.set_provider(pid)
    return adapter


def test_default_scope_requires_author_mode(tmp_path, monkeypatch):
    adapter = _adapter_with_default(tmp_path, monkeypatch, author_mode=False)
    assert not adapter.can_write("default")
    with pytest.raises(PermissionError):
        adapter.set_value("alpha", "default", 1)
    adapter.author_mode = True
    assert adapter.can_write("default")
    adapter.set_value("alpha", "default", 1)
    val, src = adapter.effective_for_key("alpha")
    assert val == 1 and src == "default"
    adapter.clear_value("alpha", "default")
    val, src = adapter.effective_for_key("alpha")
    assert val is None and src is None


def test_upsert_field_parses_default(tmp_path, monkeypatch):
    _adapter_with_default(tmp_path, monkeypatch, author_mode=True)
    author = AuthorAdapter("demo-auth")
    with pytest.raises(ValidationError):
        author.upsert_field("beta", "integer", default="1")
    author.upsert_field("beta", "integer", default=parse_field_value("integer", "1"))
    assert author.default_for_key("beta") == 1



import pytest

import pysigil.discovery as discovery
from pathlib import PurePosixPath


@pytest.mark.parametrize(
    'name,expected',
    [
        ('My_Plugin', 'my-plugin'),
        ('  fancy.PROVIDER  ', 'fancy-provider'),
        ('Mixed-._Case', 'mixed-case'),
    ],
)
def test_pep503_name(name, expected):
    assert discovery.pep503_name(name) == expected


def test_iter_installed_providers_dedup(monkeypatch):
    class DummyDist:
        def __init__(self, name, files, meta_name=None):
            self.name = name
            self._files = [PurePosixPath(p) for p in files]
            if meta_name is None:
                self.metadata = {}
            else:
                self.metadata = {"Name": meta_name}

        @property
        def files(self):
            return self._files

    dists = [
        DummyDist('pkg_a', ['.sigil/metadata.ini'], 'Foo'),
        DummyDist('pkg_b', ['.sigil/metadata.ini'], 'Foo'),  # duplicate provider id
        DummyDist('pkg_c', ['.sigil/metadata.ini'], 'Bar'),
        DummyDist('pkg_d', ['other.txt'], 'Baz'),  # not a provider
    ]

    monkeypatch.setattr(discovery, 'distributions', lambda: dists)

    providers = list(discovery.iter_installed_providers())
    assert [(pid, name) for pid, name, _ in providers] == [
        ('foo', 'Foo'),
        ('bar', 'Bar'),
    ]
